#! /usr/bin/env python3
# -*- coding: utf-8 -*-


"""
Granef API to perform Dgraph queries and provide responses with defined layout.

The default configuration exposes API at 127.0.0.1:7000. To access Swagger documentation visit
http://127.0.0.1:7000/docs.

Usage:  $ python3 main.py
        $ python3 main.py --ip "172.23.79.229"
"""

import argparse
import ujson as json  # Fast JSON parser
import re
import sys

# Modules required to run FastAPI
import uvicorn  # Python web server
from typing import Optional
from fastapi import FastAPI, Request, HTTPException
from fastapi.middleware.cors import CORSMiddleware
from pydantic import BaseModel

# Custom modules of Granef API
from dgraph_client import DgraphClient
import models
import network_queries
import host_queries
import connections_queries
import other_queries


# Application definition ("description" key may be added too).
app = FastAPI(
    title="Granef API",
    version="0.2",
)

# Load API routers
app.include_router(network_queries.router, prefix="/network_queries", tags=["network queries"])
app.include_router(host_queries.router, prefix="/host_queries", tags=["host queries"])
app.include_router(connections_queries.router, prefix="/connections_queries", tags=["connections queries"])
app.include_router(other_queries.router)


@app.get("/", summary="Get API information")
def get_root(request: Request):
    """
    Default function to show Granef API name, version, and Swagger URL when root path is requested.
    """
    swagger_path = "http://{hostname}:{port}/docs".format(hostname=request.url.hostname, port=request.url.port)
    return {"name": app.title, "version": app.version, "swagger": swagger_path}


@app.get("/connect", summary="Establish connection to Dgraph database")
def dgraph_connect():
    """
    Establish connection to Dgraph database server. When API starts the connection is automatically
    established. Call this function onl if some connection error occurred.
    """
    dgraph_client = DgraphClient()
    try:
        dgraph_client.connect(ip=args.dgraph_ip, port=args.dgraph_port)
    except Exception as e:
        raise HTTPException(
            status_code=503,
            detail=str(e)
        )
    return {"detail": "Granef API connected to Dgraph server."}


if __name__ == "__main__":
    # Argument parser automatically creates -h argument
    parser = argparse.ArgumentParser()
    parser.add_argument("--ip", help="IP address to bind the API web server.", type=str, default="0.0.0.0", required=False)
    parser.add_argument("--port", help="Port to bind the API web server.", type=int, default=7000, required=False)
    parser.add_argument("--dgraph_ip", help="Dgraph server IP addres.", type=str, default="alpha", required=False)
    parser.add_argument("--dgraph_port", help="Dgraph server port.", type=int, default=9080, required=False)
    global args
    args = parser.parse_args()

    print(f'DEBUG: main.py got args: ip={args.ip}; port={args.port}; dgraph_ip={args.dgraph_ip}; dgraph_port={args.dgraph_port}.')

    # Set HTTP headers and allow all connection
    app.add_middleware(
        CORSMiddleware,
        allow_origins=["*"],
        allow_credentials=True,
        allow_methods=["*"],
        allow_headers=["*"],
    )

    # Initialize dgraph client
    dgraph_client = DgraphClient()
    dgraph_client.connect(ip=args.dgraph_ip, port=args.dgraph_port)

    # Start API web server using Uvicorn server
    uvicorn.run(app, host=args.ip, port=args.port)
