#! /usr/bin/env python3
# -*- coding: utf-8 -*-

import sys
import os
import re
import subprocess
import argparse
import ipaddress


"""
Runs web and API services, sets env vars. 

Usage:
Independent use:
    $ service_arg_handler.py -im -ip <ip> -p <port> -di <dgraph_ip> -dp <dgraph_port>

Integrated to Orchestrator:
    $ service_arg_handler.py
    where environment variable <PREFIX>_ENDPOINT_HANDLING_ALPHA_GRPC_EXTERNAL has to be set

"""


def is_string(val):
    """
        Checks if input val is a string.

        :return: True if val is a string
    """
    return isinstance(val, str)


def is_int(val):
    """
        Checks if input val is an integer.

        :return: True if val is a integer
    """
    try:
        int(val)
        return True
    except ValueError:
        return False



def is_ip_address(string):
    """
        Check if the given string is a valid IP address.

        :param string: String that should be verified if it is IP address
        :return: True if the string is a valid IP address
    """
    try:
        ipaddress.ip_address(string)
        return True
    except ValueError:
        return False


def is_hostname(string):
    """
        Check if the given string is a valid hostname.

        Source: https://stackoverflow.com/questions/2532053/validate-a-hostname-string

        :param string: String that should be verified if it is a hostname
        :return: True if the string is a valid hostname
    """
    if len(string) > 255:
        return False
    if string[-1] == ".":
        string = string[:-1]
    allowed = re.compile("(?!-)[A-Z\d-]{1,63}(?<!-)$", re.IGNORECASE)
    return all(allowed.match(x) for x in string.split("."))


def ip_address_or_hostname(string):
    """
        Check if the given string is a valid IP address or hostname.

        :param string: String that should be verified if it is an IP address or hostname
        :return: Given string if True
    """
    return is_ip_address(string) or is_hostname(string)


def is_port(val):
    """
        Check if the given val is a valid port number.

        :param val: port number expected to be in range [0, 65536]
        :return: True if val is in valid format
    """
    return is_int(val) and 0 < int(val) < 65536


def is_ip_and_port(string):
    """
        Check if the given string is in valid format of 'ip:port' or 'hostname:port'.

        :param string: string that contains a : (e.g. '121.11.10.9:1234')
        :return: True if the string is in valid format
    """
    parts = string.split(":")

    if len(parts) != 2:
        return False

    return ip_address_or_hostname(parts[0]) or is_port(parts[1])


def extract_ip_port(val):
    parts = val.split(":")
    return parts[0], parts[1]


def is_prefix(val):
    """
        Checks if input val is a prefix string.

        :return: True if val is a prefix
    """
    return isinstance(val, str) and "-" in val


def get_prefix():
    """
        Called if independent_mode is not set. Gets name of module in Orchestrator from MODULE_ID_SELF and transforms
        it to prefix that is used as prefix for other environment variables.
    """
    print("Extracting prefix for environment variables from module name:\n")
    module_name = get_arg_value("MODULE_NAME", "", False, is_string)
    return "PARAMETER_" + module_name.upper().replace("-", "_") + "_"


def get_arg_value(name, prefix, use_prefix, check_function):
    """
        Tries to get value of environment variable name. If the env var is defined, it checks if its value is
        in correct format. If some required variable could not be found, the script will call sys.exit(1).

        :param name: name of environment variable
        :param prefix: prefix for environment variables (as specified in Orchestrator db)
        :param use_prefix: boolean indicating that a prefix should be concatenated to name
        :param check_function: name of a function that checks if the value of env var is in correct format
    """
    name = prefix + name if use_prefix else name
    if name in os.environ:
        value = os.getenv(name)
        if check_function(value):
            print("OK      {:<32} {}".format(name, value))
            return value
        else:
            print("ERROR   {:<32} {} - is not in correct format".format(name, value))
    else:
        print("ERROR   {:<32} is not defined or could not be read".format(name))
        # if this is printed check module definition on orchestrator database, there
        # should always be at least some default value returned

    print("\nExiting.")
    sys.exit(1)
    # something went wrong (either the environment variable value could not be read or it was in incorrect format)
    # therefore we do not want to continue with execution of this script - we consider each variable to be required (if
    # both of the conditions are not satisfied, it indicates some problem in communication with the database etc.)


def resolve_env_names(prefix):
    """
        Loads all environment variables to a dictionary { env_var_name : env_val_value } and checks
        whether their values are in expected format. If any value is either missing or in incorrect format the
        execution of this script is terminated.

        :param prefix: prefix for environment variables (as specified in Orchestrator db) - for easier use of arguments
                       in main
    """
    args_dict = {}

    # arg_names contains similar information to argparse.ArgumentParser() for each env var
    arg_names = [("ENDPOINT_HANDLING_ALPHA_GRPC_EXTERNAL", is_ip_and_port, False)]

    print("Loading environment variables:\n\nStatus  Name                             Value")
    for name, check_function, use_prefix in arg_names:
        value = get_arg_value(name, prefix, use_prefix, check_function)
        if value:
            args_dict[name] = value
        else:
            args_dict[name] = ""
    print("\nSuccessfully loaded environment variables.\n")

    return args_dict


def define_parsed_arguments():
    """
        Add arguments to ArgumentParser (argparse) module instance.

        Argument -m (--independent_mode) is a switch that indicates that this script will run in an independent mode
        (not integrated in the Orchestrator). The arguments are then expected to be passed as command line arguments,
        not environment variables.

        :return: parsed arguments
    """
    parser = argparse.ArgumentParser()

    parser.add_argument("-im", "--independent_mode", help="Use independent mode (not integrated to the Orchestrator)",
                        action="store_true")
    parser.add_argument("-ip", "--ip", help="IP address to bind the API web server", default="0.0.0.0",
                        required=False)
    parser.add_argument("-p", "--port", help="Port to bind the API web server", default=7000, required=False)
    parser.add_argument("-di", "--dgraph_ip_port", help="Dgraph server IP:port address", default="alpha:9080",
                        required=False)

    return parser.parse_args()


def check_and_load_parsed_arg(args_dict, script_name, arguments, arg_name, check_func, msg):
    """
        Check if argument with name arg_name from argparse module instance arguments is in the correct format according
        to check_func. If yes - then store its value in dictionary args_dict under name script_name. If not - print out
        msg and raise ArgumentTypeError.

        :param args_dict: arguments dictionary used later in main
        :param script_name: name of an argument used in this script
        :param arguments: argparse module instance with command line argument values
        :param arg_name: name of concrete argument from argparse
        :param check_func: what function should is used for value check of argument arg_name
        :param msg: what msg will be printed out if the value of argument is incorrect
    """
    value = getattr(arguments, arg_name)
    if check_func(value):
        args_dict[script_name] = value
    else:
        raise argparse.ArgumentTypeError(msg + value)


def resolve_parsed_names(arguments):
    """
        Helper function to make the use of environment variables and command line arguments easier in main.

        :param arguments: argparse module instance with command line argument values
        :return: dictionary with converted command line arguments names and their values
    """
    args_dict = {}
    check_and_load_parsed_arg(args_dict, "ENDPOINT_HANDLING_ALPHA_GRPC_EXTERNAL", arguments, "dgraph_ip_port",
                              is_ip_and_port, "given string is not a valid ip:port: ")
    return args_dict


if __name__ == "__main__":
    args_loaded = define_parsed_arguments()

    if getattr(args_loaded, "independent_mode"):
        args = resolve_parsed_names(args_loaded)
    else:
        prefix = get_prefix()
        args = resolve_env_names(prefix)
        # this environment variables prefix corresponds to the module name in orchestrator db

    ip = str(args_loaded.ip)
    port = str(args_loaded.port)  # 7000 is default, in orchestrator can't be changed
    dgraph_ip_port = str(args["ENDPOINT_HANDLING_ALPHA_GRPC_EXTERNAL"])
    dgraph_ip, dgraph_port = extract_ip_port(dgraph_ip_port)

    # definition of API service in supervisor.conf expects arguments to be passed as hardcoded env vars,
    # set them according to arguments for this script:
    os.environ["IP"] = ip
    os.environ["PORT"] = port
    os.environ["DGRAPH_IP"] = dgraph_ip
    os.environ["DGRAPH_PORT"] = dgraph_port
    print(f"Set values of env vars IP='{ip}', PORT='{port}', DGRAPH_IP='{dgraph_ip}' and DGRAPH_PORT='{dgraph_port}'.")

    try:
        services_command = "supervisord -c /etc/supervisor.conf"
        subprocess.run(services_command, shell=True, check=True)

    except subprocess.CalledProcessError as e:
        print("Error (CalledProcessError) when running supervisord:\nCommand =" + str(e.cmd) + ",\nReturncode =" +
              str(e.returncode) + ",\nOutput =" + str(e.output))
    except Exception as e:
        print("Error when running supervisord: " + str(e))
