#! /usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Script:
1. Loads PCAP file (or zipped PCAP) from REST API or uses a local directory.
2. The file (if there is any) is then transformed using Zeek to logs.
3. The result is uploaded back to REST API or a local directory.

Usage:=
Independent use:
* Obtain data via REST API:
    $ communication_handler.py -im -i <pcap_file_name> -ip <API address:port>

* Obtain data via local directory:
    $ communication_handler.py -im -l -i <pcap_file_name> -in <input_dir> -ou <output_dir>


Integrated to Orchestrator:
* Obtain data via REST API:
    $ python3 communication_handler.py
     where environment variables PARAMETER_<PREFIX>_LOCAL=False, DATAWAREHOUSE, PARAMETER_<PREFIX>_INPUT_DIR, 
     PARAMETER_<PREFIX>_OUTPUT_DIR, PARAMETER_<PREFIX>_ZIP_DIR and PARAMETER_<PREFIX>_INPUT_IDENTIFIER have to be set

* Obtain data via local directory:
    $ python3 communication_handler.py
     where environment variables PARAMETER_<PREFIX>_LOCAL=True, PARAMETER_<PREFIX>_INPUT_DIR, 
     PARAMETER_<PREFIX>_OUTPUT_DIR and PARAMETER_<PREFIX>_INPUT_IDENTIFIER
     have to be set
     (DATAWAREHOUSE has to be arbitrary correct ip:port just to pass its value format check, but it will
     not be used)
"""

import sys
import re
import fnmatch
import os
import subprocess
import argparse
import ipaddress
from pathlib import Path
sys.path.insert(0, str(Path(Path(__file__).parent / Path('python-analysis-modules')).resolve()))
import python_analysis_modules.warehouse_connector as warehouse_connector
import python_analysis_modules.zip_functions as zip_functions


def is_string(val):
    """
        Checks if input val is a string.

        :return: True if val is a string
    """
    return isinstance(val, str)


def is_dir_path(string):
    """
        Check if the given string is a path to an existing directory.

        :param string: String that should be verified if it is a directory
        :return: True if given string is a path to an existing directory
    """
    return os.path.isdir(string)


def is_abs_dir_path(string):
    """
        Checks if string is an absolute path to an existing directory.

        :return: True if given string is an absolute path to an existing directory
    """
    return len(string) > 0 and string[0] == '/' and is_dir_path(string)


def is_pcap_file(string):
    """
        Check if the given string end with '.pcap'.

        :param string: String that should be verified if it is a PCAP file
        :return: True if given val is in correct format
    """
    return fnmatch.fnmatch(string, '*.pcap')


def is_input_file(string):
    """
        Check if the given string end with '.pcap' or is a string (uuid).

        :param string: String that should be verified if it is in the correct format
        :return: True if given val is in correct format
    """
    return is_pcap_file(string) or is_string(string)


def is_ip_address(string):
    """
        Check if the given string is a valid IP address.

        :param string: String that should be verified if it is IP address
        :return: True if the string is a valid IP address
    """
    try:
        ipaddress.ip_address(string)
        return True
    except ValueError:
        return False


def is_hostname(string):
    """
        Check if the given string is a valid hostname.

        Source: https://stackoverflow.com/questions/2532053/validate-a-hostname-string

        :param string: String that should be verified if it is a hostname
        :return: True if the string is a valid hostname
    """
    if len(string) > 255:
        return False
    if string[-1] == ".":
        string = string[:-1]
    allowed = re.compile("(?!-)[A-Z\d-]{1,63}(?<!-)$", re.IGNORECASE)
    return all(allowed.match(x) for x in string.split("."))


def ip_address_or_hostname(string):
    """
        Check if the given string is a valid IP address or hostname.

        :param string: String that should be verified if it is an IP address or hostname
        :return: Given string if True
    """
    return is_ip_address(string) or is_hostname(string)


def is_port(val):
    """
        Check if the given val is a valid port number.

        :param val: port number expected to be in range [0, 65536]
        :return: True if val is in valid format
    """
    return is_integer(val) and 0 < val < 65536


def is_ip_and_port(string):
    """
        Check if the given string is in valid format of 'ip:port' or 'hostname:port'.

        :param string: string that contains API address : (e.g. '121.11.10.9:1234')
        :return: True if the string is in valid format
    """
    parts = string.split(":")

    if len(parts) != 2:
        return False

    return ip_address_or_hostname(parts[0]) or is_port(parts[1])


def is_boolean(val):
    """
        Check if the given value is a bool.

        :param val: input value
        :return: True if it is an bool
    """
    try:
        eval(val)
        return True
    except NameError:
        # conversion to bool was not successful
        return False


def is_prefix(val):
    """
        Checks if input val is a prefix string.

        :return: True if val is a prefix
    """
    return isinstance(val, str) and '-' in val


def check_input_availability(dir_path, input_name):
    """
        Check file availability in input_dir directory if the "local" argument is set to True.

        :param dir_path: path to input directory
        :param input_name: name of input file
        :return: local input file path
    """
    file_path = dir_path + '/' + input_name
    if os.path.isfile(file_path):
        return file_path
    else:
        print('ERROR file ' + file_path + ' not found')
        return ''


def get_input(local, dir_path, input_name, warehouse_connector):
    """
        Either check input availability in local storage and correct format (PCAP), or use warehouse_connector to
        get input and then unzip it and then check the format of unzipped file(s).

        :param local: boolean that indicated that the input file is in local storage
        :param dir_path: path where the input should be located/ downloaded to
        :param input_name: name or uuid of input file
        :param warehouse_connector: WarehouseConnector instance
        :return: path to input file
    """
    if local:
        downloaded_file = check_input_availability(dir_path, input_name)
        return downloaded_file if is_pcap_file(downloaded_file) else ''
    else:
        downloaded_file = warehouse_connector.download_input()

    if downloaded_file and not is_pcap_file(downloaded_file):
        # extract downloaded_file to input_dir (skip if its not a zip file):
        extracted_dir_path = zip_functions.extract_file(downloaded_file, dir_path)
        # get extracted pcap file name (skip other files in extracted_dir_path):
        # if more than one file was found here, a list is returned instead
        downloaded_file = zip_functions.get_filenames_in_dir(extracted_dir_path)

    if is_string(downloaded_file) and is_pcap_file(downloaded_file):
        return downloaded_file
    else:
        print('ERROR Input file ' + input_name + ' was found (and extracted), but does not contain a PCAP file.')

    return ''


def upload_zipped_output(output_dir, zip_dir, warehouse_connector):
    """
        Zips generated output log files and uploades to the Data Warehouse using warehouse_connector.

        :param output_dir: path to directory containing output log files
        :param zip_dir: absolute path to a directory where the output zip file will be saved
        :param warehouse_connector: WarehouseConnector instance
    """
    # zip all output log files to a single output file:
    zip_name = wc.get_output_name()
    if zip_functions.zip_files_in_dir(output_dir, zip_dir, zip_name, ['.log']):
        warehouse_connector.upload_output()
    else:
        print('ERROR file (' + zip_name + ') containing the output logs could not be zipped.')


def get_prefix():
    """
        Called if independent_mode is not set. Gets name of module in Orchestrator from MODULE_ID_SELF and transforms
        it to prefix that is used as prefix for other environment variables.
    """
    print('Extracting prefix for environment variables from module name:\n')
    module_name = get_arg_value('MODULE_NAME', '', False, is_string)
    return 'PARAMETER_' + module_name.upper().replace('-', '_') + '_'


def get_arg_value(name, prefix, use_prefix, check_function):
    """
        Tries to get value of environment variable name. If the env var is defined, it checks if its value is
        in correct format. If some required variable could not be found, the script will call sys.exit(1).

        :param name: name of environment variable
        :param prefix: prefix for environment variables (as specified in Orchestrator db)
        :param use_prefix: boolean indicating that a prefix should be concatenated to name
        :param check_function: name of a function that checks if the value of env var is in correct format
    """
    name = prefix + name if use_prefix else name
    if name in os.environ:
        value = os.getenv(name)
        if check_function(value):
            print('OK      {:<32} {}'.format(name, value))
            return value
        else:
            print('ERROR   {:<32} {} - is not in correct format'.format(name, value))
    else:
        print('ERROR   {:<32} is not defined or could not be read'.format(name))
        # if this is printed check module definition on orchestrator database, there
        # should always be at least some default value returned

    print('\nExiting.')
    sys.exit(1)
    # something went wrong (either the environment variable value could not be read or it was in incorrect format)
    # therefore we do not want to continue with execution of this script - we consider each variable to be required (if
    # both of the conditions are not satisfied, it indicates some problem in communication with the database etc.)


def resolve_env_names(prefix):
    """
        Loads all environment variables to a dictionary { env_var_name : env_val_value } and checks
        whether their values are in expected format. If any value is either missing or in incorrect format the
        execution of this script is terminated.

        :param prefix: prefix for environment variables (as specified in Orchestrator db) - for easier use of arguments
                       in main
    """
    args_dict = {}

    # arg_names contains similar information to argparse.ArgumentParser() for each env var
    arg_names = [('LOCAL', is_boolean, True),
                 ('DATAWAREHOUSE', is_ip_and_port, False),
                 ('INPUT_DIR', is_dir_path, True),
                 ('OUTPUT_DIR', is_dir_path, True),
                 ('ZIP_DIR', is_abs_dir_path, True),
                 ('INPUT_IDENTIFIER', is_input_file, True)]

    print('Loading environment variables:\n\nStatus  Name                             Value')
    for name, check_function, use_prefix in arg_names:
        value = get_arg_value(name, prefix, use_prefix, check_function)
        if value:
            args_dict[name] = value
        else:
            args_dict[name] = ''
    print('\nSuccessfully loaded environment variables.\n')

    return args_dict


def define_parsed_arguments():
    """
        Add arguments to ArgumentParser (argparse) module instance.

        Argument -m (--independent_mode) is a switch that indicates that this script will run in an independent mode
        (not integrated in the Orchestrator). The arguments are then expected to be passed as command line arguments,
        not environment variables.

        :return: parsed arguments
    """
    parser = argparse.ArgumentParser()

    parser.add_argument('-im', '--independent_mode', help='Use independent mode (not integrated to the Orchestrator)',
                        action='store_true')
    parser.add_argument('-in', '--input_dir', help='Input data directory', default='/usr/local/bin/granef/data/input/')
    parser.add_argument('-ou', '--output_dir', help='Output data directory',
                        default="/usr/local/bin/granef/data/output/")
    parser.add_argument('-zd', '--zip_dir_path', help='Path to output zip directory',
                        default="/usr/local/bin/granef/data/zip/")
    parser.add_argument('-i', '--input_identifier', help='Input PCAP file name/ uuid', required='-im' in sys.argv
                                                                                                or '--independent_mode' in sys.argv)
    parser.add_argument('-ip', '--ip_and_port', help='IP address and port of REST API', default='127.0.0.1:5000')
    parser.add_argument('-l', '--local', help='Use a local directory instead of REST API', action='store_true')

    return parser.parse_args()


def check_and_load_parsed_arg(args_dict, script_name, arguments, arg_name, check_func, msg):
    """
        Check if argument with name arg_name from argparse module instance arguments is in the correct format according
        to check_func. If yes - then store its value in dictionary args_dict under name script_name. If not - print out
        msg and raise ArgumentTypeError.

        :param args_dict: arguments dictionary used later in main
        :param script_name: name of an argument used in this script
        :param arguments: argparse module instance with command line argument values
        :param arg_name: name of concrete argument from argparse
        :param check_func: what function should is used for value check of argument arg_name
        :param msg: what msg will be printed out if the value of argument is incorrect
    """
    value = getattr(arguments, arg_name)
    if check_func(value):
        args_dict[script_name] = value
    else:
        raise argparse.ArgumentTypeError(msg + value)


def resolve_parsed_names(arguments):
    """
        Helper function to make the use of environment variables and command line arguments easier in main.

        :param arguments: argparse module instance with command line argument values
        :return: dictionary with converted command line arguments names and their values
    """
    args_dict = {}
    args_dict['LOCAL'] = str(getattr(arguments, 'local'))
    check_and_load_parsed_arg(args_dict, 'DATAWAREHOUSE', arguments, 'ip_and_port', is_ip_and_port,
                              'given string is not a valid <ip:port>: ')
    check_and_load_parsed_arg(args_dict, 'INPUT_DIR', arguments, 'input_dir', is_dir_path,
                              'given string is not a valid path to an existing directory: ')
    check_and_load_parsed_arg(args_dict, 'OUTPUT_DIR', arguments, 'output_dir', is_dir_path,
                              'given string is not a valid path to an existing directory: ')
    check_and_load_parsed_arg(args_dict, 'ZIP_DIR', arguments, 'zip_dir_path', is_abs_dir_path,
                              'given string is not a valid absolute path to an existing directory: ')
    check_and_load_parsed_arg(args_dict, 'INPUT_IDENTIFIER', arguments, 'input_identifier', is_input_file,
                              'given string is not a valid path to a PCAP file or uuid: ')
    return args_dict


if __name__ == '__main__':
    args_loaded = define_parsed_arguments()

    if getattr(args_loaded, 'independent_mode'):
        args = resolve_parsed_names(args_loaded)
    else:
        prefix = get_prefix()
        args = resolve_env_names(prefix)
        # this environment variables prefix corresponds to the module name in orchestrator db

    use_local = eval(args['LOCAL'])
    input_dir = args['INPUT_DIR']
    output_dir = args['OUTPUT_DIR']
    zip_dir = args['ZIP_DIR']
    pcap_identifier = args['INPUT_IDENTIFIER']

    wc = warehouse_connector.WarehouseConnector(pcap_identifier,
                                                input_dir,
                                                zip_dir,
                                                'extracted',
                                                False,
                                                'zeek_extraction',
                                                None,
                                                None,
                                                None,
                                                'http://' + args['DATAWAREHOUSE'] + '/')

    input_pcap = get_input(use_local, input_dir, pcap_identifier, wc)

    if input_pcap:
        try:
            zeek_command = '/opt/zeek/bin/zeek -Cr ' + input_pcap + ' local.zeek "' + \
                           'Site::local_nets += {192.168.0.0/16, 172.16.0.0/12, 10.0.0.0/8, [fe80::]/10}; ' + \
                           'redef LogAscii::use_json=T; ' + \
                           'redef LogAscii::json_timestamps = JSON::TS_ISO8601;"'
            subprocess.run(zeek_command, shell=True, capture_output=True, check=True, cwd=output_dir)

            if not use_local:
                upload_zipped_output(output_dir, zip_dir, wc)

        except subprocess.CalledProcessError as e:
            print('Error (CalledProcessError) when running zeek:\nCommand =' + str(e.cmd) + ',\nReturncode =' +
                  str(e.returncode) + ',\nOutput =' + str(e.output))
        except Exception as e:
            print('Error when running zeek: ' + str(e))

    else:
        print('No PCAP files were loaded. No log files generated.')
