#! /usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Class handles communication of Analytical Modules with Data Warehouse.

Analytical Modules functionality from the Data Warehouses perspective:

  | input | --> | module | --> | output |

Each object in Warehouse is considered to be a DataFile. There will be created Links between these DataFiles
named derivedFrom and some other details will be stored as a GenericObject called DerivationDetail. CaseId for the
new objects is set to the same caseId of the input file.

Example for Data Extraction:

  | DataFile containing zipped pcap file | <-- derivedFrom -- | DataFile containing zipped Zeek log files |
                                                                                    |
                                                                            derivationOutput
                                                                                    |
                                                                                    v
                                                                  | GenericObject with type DerivationDetail |
                                                               (name derivation_string = e.g. 'zeek_extraction')
                                                                                    |
                                                                                    v
                                                                        | UserProperty objects |
                                                                   (optionally some module arguments)

    Strings that are currently being used as derivation_string for Analytical Modules:
      extraction = zeek_extraction
      transformation = dgraph_transformation
      indexing = dgraph_indexing

Optimization: Analytical Module will first check if it hasn't already processed the given input. This can be concluded
by existence of particular info in DerivationDetail object.


Usage:=
    import warehouse_connector

    # 1. create a WarehouseConnector instance:
    wc = warehouse_connector.WarehouseConnector(<input_identifier>,
                                                <input_dir>,
                                                <output_dir>,
                                                <zip_suffix>,
                                                <suffix_used>,
                                                <derivation_string>,
                                                <previous_derivation_string>,
                                                <arg_names>,
                                                <module_arg_dict>,
                                                <address>)

    # concrete example:
    wc = warehouse_connector.WarehouseConnector('0f75b79b-b4fd-471b-8bd4-8ecbffc79f2b',
                                                '/usr/local/bin/granef/data/input/',
                                                '/usr/local/bin/granef/data/zip/',
                                                        # yes, modules use 'zip' here instead of 'output' - because
                                                        # there will be all output files zipped into a single file
                                                'extracted',
                                                False,
                                                'zeek_extracted',
                                                None,
                                                None,   # or e.g. ['ENDPOINTS_WAREHOUSE','INPUT_DIR']
                                                None,   # or a dictionary containing all modules arguments
                                                'http://147.251.124.63:8080/')

    # 2. download input:
    downloaded_file = wc.download_input()  # then unzip using zip_functions.py

    # 3. get output zip name:
    zip_name = wc.get_output_name()  # then zip

    # 4. upload output (WarehouseConnector expects zipped file under name self.__output_name in self.__output_dir):
    wc.upload_output()
"""

import os
import ujson as json
import requests


class WarehouseConnector:
    def __init__(self, input_identifier, input_dir, output_dir, zip_suffix, suffix_used, derivation_string,
                 previous_derivation_string, arg_names, module_arg_dict, address):
        """
        Class constructor.

        :param input_identifier: uuid of input file
        :param input_dir: path to input data directory where the file should be downloaded to
        :param output_dir: path to a directory where will be the final output
        :param zip_suffix: suffix of uploaded output file name
        :param suffix_used: boolean - True if suffix was already used before and is supposed to be removed
        :param derivation_string: string that is stored as name in DerivationDetail and is used to check if input
               was already processed
        :param previous_derivation_string: string that was stored as name in DerivationDetail for file generated by
               previous Analytical Module (None is used for extraction module)
        :param arg_names: names of module arguments that will be stored as UserProperty in Data Warehouse
        :param module_arg_dict: dictionary containing module arguments (e.g. input_dir -> '/home')
        :param address: address of REST API
        """
        # check that arg_names and module_arg_dict are valid:
        if (arg_names is not None and module_arg_dict is None) or (arg_names is None and module_arg_dict is not None):
            raise RuntimeWarning('Both self.__arg_names, self.__module_arg_dict are supposed to be None or neither.')

        if arg_names is not None and module_arg_dict is not None:
            for arg_name in arg_names:
                if arg_name not in module_arg_dict:
                    raise RuntimeWarning('Given arg key "' + arg_name + '" is not a valid key in module_arg_dict.')

        self.__input_name = ''
        self.__input_uuid = input_identifier
        self.__input_dir = input_dir

        self.__output_name = ''
        self.__output_uuid = ''
        self.__output_dir = output_dir

        self.__zip_suffix = zip_suffix
        self.__suffix_used = suffix_used

        self.__case_id = ''

        self.__derivation_string = derivation_string
        self.__previous_derivation_string = previous_derivation_string
        self.__details_uuid = ''
        self.__arg_names = arg_names
        self.__module_arg_dict = module_arg_dict

        self.__address = address

    def get_output_name(self):
        return self.__output_name

    def get_case_id(self):
        return self.__case_id

    def __json_contains_check(self, key, json_metadata, obj_string):
        """
            Check if given json_metadata contains all keys from key_words. If not, ValueError exception is raised.

            :param key: key string
            :param json_metadata: objects metadata in JSON format
            :param obj_string: some string that will be printed out in case of missing key to help identify the source
                   of error
            :return True if key is in json_metadata
        """
        print('Checking if ' + obj_string + ' contains optional key "' + key + '" in JSON metadata. Exiting.')
        return key in json_metadata

    def __json_contains_or_skip(self, key_words, json_metadata, obj_string):
        """
            Check if given json_metadata contains all keys from key_words. If not, ValueError exception is raised.

            :param key_words: list containing keys
            :param json_metadata: objects metadata in JSON format
            :param obj_string: some string that will be printed out in case of missing key to help identify the source
                   of error
        """
        for key in key_words:
            if key not in json_metadata:
                raise ValueError('ERROR Object ' + obj_string + ' does not contain some necessary key [' +
                                 ', '.join(map(str, key_words)) + '] in JSON metadata. Exiting.')

    def __details_processed(self, json_details):
        """
            More complex check if input was already processed - consider also provided argument (self.__arg_names)
            values in self.__module_arg_dict.

            :param json_details: saved DerivationDetail object in Data Warehouse
            :return: True if all UserProperty in json_details (DerivationDetail object) have same values as those
                     in self.__module_arg_dict
        """
        # if no arguments were given, it means that I don't have to check them in order to stop Analytical Module from
        # running again (I only care about self.__derivation_string):
        if self.__arg_names is None and self.__module_arg_dict is None:
            return True

        # both are not None - check all saved UserProperty objects against all self.__arg_names values:
        properties_counter = len(self.__arg_names)
        self.__json_contains_or_skip(['properties'], json_details, ' DerivationDetail linked to input')
        properties = json_details['properties']

        for property in properties:
            self.__json_contains_or_skip(['propertyKey'], property, ' property in DerivationDetail linked to input')
            property_key = property['propertyKey']
            # , 'valueString', 'valueLong', 'valueDouble'

            # convert saved 'propertyKey' name to format expected by self.__module_arg_dict:
            if property_key.startswith('arg_'):
                arg_name = property_key[4:].upper()
                if arg_name in self.__module_arg_dict and (
                    (self.__json_contains_check('valueString', property,
                                                ' property in DerivationDetail linked to input')
                     and property['valueString'] == str(self.__module_arg_dict[arg_name]))
                    or (self.__json_contains_check('valueLong', property,
                                                   ' property in DerivationDetail linked to input')
                        and property['valueLong'] == int(self.__module_arg_dict[arg_name]))
                    or (self.__json_contains_check('valueDouble', property,
                                                   ' property in DerivationDetail linked to input')
                        and property['valueDouble'] == float(self.__module_arg_dict[arg_name]))):
                    properties_counter -= 1  # one argument value was already used before

        # all values of saved arg values are the same as the ones given to the Analytical Module right now => skip:
        return properties_counter == 0

    def __input_already_processed(self):
        """
            Check if a link with specified DerivationDetail from object identified by its self.__input_uuid already
            exists in Data Warehouse. It prevents the Analytical Module from running again on the same data it has
            already processed before.

            :return: True if link already exists
        """
        try:
            derivation_outputs = requests.get(self.__address + '/generic-objects/derivation/' + self.__input_uuid)
            print('Response code for DerivationDetail of object with with uuid "' + self.__input_uuid + '": ' +
                  str(derivation_outputs.status_code))

            if derivation_outputs:
                metadata = derivation_outputs.text
                json_metadata = json.loads(metadata)
                for derivation in json_metadata:
                    self.__json_contains_or_skip(['name'], derivation, 'DerivationDetail linked to input')
                    if derivation['name'] == self.__derivation_string and self.__details_processed(derivation):
                        return True

        except requests.exceptions.RequestException as err:
            print('ERROR from requests module when downloading (RequestException):', err)
        except Exception as err:
            print('ERROR when downloading:', err)

        return False

    def __generate_output_name(self, ending):
        name = str(os.path.splitext(self.__input_name)[0])

        if self.__suffix_used:
            name = name.rsplit('-', 1)[0]
            print(input)

        name += '-' + self.__zip_suffix + ending
        return name

    def __get_input_content(self):
        """
            Get input files content.

            :return: path to downloaded file
        """
        file_data = requests.get(self.__address + 'generic-objects/data-files/' + self.__input_uuid, stream=True)
        print('Response code for file data "' + self.__input_name + '": ' + str(file_data.status_code))
        file_path = self.__input_dir + '/' + self.__input_name  # where the requested file will be downloaded to

        if file_data:
            with open(file_path, 'wb') as f:
                # use chunks because the file might be too big to fit in RAM:
                for chunk in file_data.iter_content(32 * 1024):
                    f.write(chunk)

            # create self.__output_name from downloaded files name:
            self.__output_name = self.__generate_output_name('.zip')
            return file_path

        return ''

    def __check_type(self, link, type):
        self.__json_contains_or_skip(['type'], link, 'type of link')
        link_type = link['type']
        return link_type == type

    def __pseudo_dfs(self, uuid, link_type, direction, visited):
        """
            Pseudo DFS used to find the corresponding input for current Analytical Module. Actual input can be
            identified DerivationDetail name == self.__previous_derivation_string.

            Pseudo-DFS is based on db schema:
            input uuid -> for derivedFrom links --> left DataFile uuid --> for derivationOutput links -->
            DerivationDetail name check --> right DataFile uuid

            :param uuid: uuid of input DataFile
            :param link_type: string 'derivedFrom' or 'derivationOutput' specifies the type of searched links
            :param direction: string 'left' if link_type is 'derivedFrom', 'right' if 'derivationOutput', specifies
                   the direction of depth search (it's not necessary to go in both ways)
            :param visited: uuid of visited links
            :return: True if file found
        """
        # go over all links with specified link_type ('derivedFrom'/ 'derivationOutput'):
        response = requests.get(self.__address + 'links/' + str('right' if direction == 'left' else 'left') + '/' +
                                uuid)
        print('Response for links (' + str(response.status_code) + '): ' + str(response.content))
        metadata = response.text
        json_metadata = json.loads(metadata)

        for link in json_metadata:
            self.__json_contains_or_skip(['uuid', direction], link, 'link from ' + uuid)
            if link['uuid'] in visited:
                continue
            visited.append(link['uuid'])

            if not self.__check_type(link, link_type):
                continue

            # in 'derivedFrom' output we will focus on 'left', in 'derivationOutput' the output can be directly found
            # under 'outputUuid' in properties of 'right':
            self.__json_contains_or_skip([direction], link, direction + ' in link from ' + uuid)
            response = requests.get(self.__address + 'generic-objects/' + link[direction], stream=True)
            print('Response for metadata_in_direction in direction ' + direction + ' (' + str(response.status_code)
                  + ') ' + ' (uuid: ' + link[direction] + ') in link from ' + uuid + ': ' + str(response.content))
            metadata_in_direction = response.text
            json_metadata_in_direction = json.loads(metadata_in_direction)
            self.__json_contains_or_skip(['uuid', 'name', 'properties'], json_metadata_in_direction, 'uuid of ' +
                                         direction + ' in link from ' + uuid)
            uuid_in_direction = json_metadata_in_direction['uuid']

            if link_type == 'derivedFrom':
                # check DerivationDetail (link_type 'derivationOutput') for self.__previous_derivation_string match:
                return self.__pseudo_dfs(uuid_in_direction, 'derivationOutput', 'right', visited)

            # check if 'name' of DerivationDetail is same as string used by previous Analytical Module:
            elif json_metadata_in_direction['name'] == self.__previous_derivation_string:
                # find 'uuid' (based on property with 'propertyKey' 'outputUuid'):
                for property in json_metadata_in_direction['properties']:
                    self.__json_contains_or_skip(['propertyKey', 'valueString'], property, 'property of right link')
                    if property['propertyKey'] == 'outputUuid':
                        self.__input_uuid = property['valueString']
                        print('Input uuid set to ' + str(property['valueString']))
                        return True
            else:
                return self.__pseudo_dfs(uuid, 'derivedFrom', 'left', visited)

        return False

    def __find_actual_input(self):
        """
            Helper method containing try catch block for self.__pseudo_dfs call on given self.__input_uuid.

            :return: True if actual input was found
        """
        try:
            print('Trying to find uuid of actual input for this Analytical Module (DFS):')
            return self.__pseudo_dfs(self.__input_uuid, 'derivedFrom', 'left', [])

        except requests.exceptions.RequestException as err:
            print('ERROR from requests module when downloading (RequestException):', err)
        except Exception as err:
            print('ERROR when downloading:', err)

        print('Actual inputs uuid not found.')
        return False

    def __post_output(self):
        """
            Post output file.

            :return: True if upload was successful
        """
        file_name = self.__output_dir + '/' + self.__output_name
        file = {'file': open(file_name, 'rb')}

        if file:
            try:
                params = {'name': self.__output_name, 'caseId': self.__case_id}
                response = requests.post(self.__address + 'generic-objects/data-files', params=params, files=file)
                print('Response for file "' + self.__output_name + '" (' + str(response.status_code) + '): ' +
                      str(response.content))

                if response:
                    # load 'uuid' of uploaded file and store it for link creation:
                    json_metadata = json.loads(response.text)
                    self.__json_contains_or_skip(['uuid'], json_metadata, 'file ' + file_name)
                    uuid = str(json_metadata['uuid'])
                    self.__output_uuid = uuid
                    return True

            except requests.exceptions.RequestException as err:
                print('ERROR from requests module when uploading:', err)
            except Exception as err:
                print('ERROR when posting zipped output:', err)
        else:
            print('ERROR output file not found ' + file)

        return False

    def __put_property_to_details(self, key, value):
        """
            Put UserProperty to DerivationDetail with uuid self.__details_uuid.

            :param key: some string (e.g. argument name, 'arg_input_dir')
            :param value: string/ number (e.g. argument value, '/home/name/input/')
            :return: True if the creation of UserProperty was successful
        """
        params = {'key': key, 'value': value}
        response = requests.put(self.__address + '/generic-objects/' + self.__details_uuid, params=params)
        print('Response for object UserProperty of DerivationDetail with uuid ' + self.__details_uuid +
              '" (' + str(response.status_code) + '): ' + str(response.content))
        return True if response else False

    def __upload_arg_details(self):
        if self.__arg_names is not None and self.__module_arg_dict is not None:
            for arg_name in self.__arg_names:
                self.__put_property_to_details('arg_' + arg_name.lower(), self.__module_arg_dict[arg_name])

    def __upload_details(self):
        """
            Create GenericObject with type DerivationDetail and name self.__derivation_string. This object is used for
            determining whether the analytical module should process input data next time it runs on self.__input_uuid
            or if it's not necessary.

            :return: True if upload was successful
        """
        try:
            # post DerivationDetail with name self.__derivation_string:
            params = {'name': self.__derivation_string, 'caseId': self.__case_id, 'type': 'DerivationDetail'}
            response = requests.post(self.__address + 'generic-objects/', params=params)
            print('Response for object "DerivationDetail ' + self.__derivation_string + '" (' +
                  str(response.status_code) + '): ' + str(response.content))

            if response:
                json_metadata = json.loads(response.text)
                self.__json_contains_or_skip(['uuid'], json_metadata, 'DerivationObject ' + self.__derivation_string)
                self.__details_uuid = str(json_metadata['uuid'])

                # put UserProperty to DerivationDetail:
                return self.__put_property_to_details('inputUuid', self.__input_uuid) and \
                       self.__put_property_to_details('outputUuid', self.__output_uuid)

        except requests.exceptions.RequestException as err:
            print('ERROR from requests module when creating DerivationDetail:', err)
        except Exception as err:
            print('ERROR DerivationDetail ' + self.__derivation_string + ' could not be created:', err)

        return False

    def __link_objects(self, left, right, type, oriented):
        """
            Create link between objects left and right (their uuid) with type 'type'.

            :param left: left objects uuid
            :param right: right objects uuid
            :param type: string
            :param oriented: boolean
        """
        try:
            params = {'left': left, 'right': right, 'type': type, 'oriented': oriented, 'caseId': self.__case_id}
            response = requests.post(self.__address + 'links', params=params)
            print('Response for link (' + str(response.status_code) + '): ' + str(response.content))

        except requests.exceptions.RequestException as err:
            print('ERROR from requests module when creating a link:', err)
        except Exception as err:
            print('ERROR link between ' + left + ' and ' + right + ' could not be created:', err)

    def download_input(self):
        """
            Use REST API to download file (specified by self.__input_uuid) to self.__input_dir directory and save
            it under its name. Tries to find actual inputs uuid if self.__previous_derivation_string is not None.

            :return: local input file path
        """
        try:
            if self.__previous_derivation_string is not None and not self.__find_actual_input():
                return ''

            if self.__input_already_processed():
                raise RuntimeWarning('File with uuid "' + self.__input_uuid + '" was already processed (there ' +
                                     'exists object DerivationDetail with name "' + self.__derivation_string +
                                     '" and the corresponding values of stored argument values). Skipping module ' +
                                     'functionality.')
            else:
                print("Input not already processed.")

            # try to get metadata of object specified by its self.input_uuid:
            file_obj = requests.get(self.__address + 'generic-objects/' + self.__input_uuid)
            print('Response code for file with uuid "' + self.__input_uuid + '": ' + str(file_obj.status_code))

            if file_obj:
                # get 'caseId' and 'name' from the objects metadata in Warehouse:
                file_metadata = file_obj.text
                json_metadata = json.loads(file_metadata)

                self.__json_contains_or_skip(['caseId', 'name'], json_metadata, self.__input_uuid)

                self.__case_id = str(json_metadata['caseId'])
                self.__input_name = str(json_metadata['name'])

                # get file content:
                return self.__get_input_content()

        except requests.exceptions.RequestException as err:
            print('ERROR from requests module when downloading (RequestException):', err)
        except Exception as err:
            print('ERROR when downloading:', err)

        return ''

    def upload_output(self):
        """
            Upload output file from self.__output_dir directory with name self.__output_name to
            Data Warehouse. If this part successful, the DerivationDetail GenericObject and links are created.
        """
        if self.__post_output():
            if self.__upload_details():
                self.__link_objects(self.__output_uuid, self.__input_uuid, 'derivedFrom', True)
                self.__link_objects(self.__output_uuid, self.__details_uuid, 'derivationOutput', False)
                self.__upload_arg_details()
            else:
                print('ERROR details of output could not be uploaded, therefore no links were created.')
        else:
            print('ERROR output could not be uploaded.')
