#! /usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Auxiliary functions for extracting and zipping files.
"""

import os
import zipfile


def zip_files_in_dir(dir, zip_dir, zip_file_name, filter):
    """
        Zips files that are in the given directory dir and the result is saved to zip_dir into a file with name
        zip_file_name (zip_dir must be an absolute path to a directory!).
        Source: https://thispointer.com/python-how-to-create-a-zip-archive-from-multiple-files-or-directory/

        :param dir: path to directory which contains files that will be zipped
        :param zip_dir: absolute path to directory where the zipped file will be saved
        :param zip_file_name: name of zipped file
        :param filter: list of endings of file names in dir that will be included in zip (e.g. '.log'), other files
                won't be included in zip
        :return: path to the zipped file
    """
    if not os.path.isdir(dir) or not os.path.isdir(zip_dir):
        return False

    prev_wdir = os.getcwd()
    zip_name = zip_dir + '/' + zip_file_name

    if not zip_name.endswith('.zip'):
        zip_name += '.zip'

    # create a ZipFile object
    with zipfile.ZipFile(zip_name, 'w') as zip_obj:
        # change working directory to the directory that is supposed to be zipped (otherwise there will also be
        # other superior folders)
        os.chdir(dir)
        # iterate over all files in working directory
        for folder_name, _, filenames in os.walk('./'):
            for filename in filenames:
                # zip only files with specified ending:
                for ending in filter:
                    if filename.endswith(ending):
                        # create complete filepath of file in directory
                        file_path = os.path.join(folder_name, filename)
                        # add file to zip_obj
                        zip_obj.write(file_path)

    # close the zip_obj object
    zip_obj.close()
    # change working directory to original working directory
    os.chdir(prev_wdir)

    return True


def extract_file(zipped_file, unzip_dir):
    """
        Extracts file with name zipped_file to unzip_dir directory. If given zipped_file is not a valid zip file,
        the extraction is skipped and zipped_file is returned (in this case it is expected to be an already extracted
        file).

        :param zipped_file: path to the file that is supposed to be extracted
        :param unzip_dir: path to a directory to which the content of zipped_file is supposed be extracted
        :return: path to the directory to which the content of zipped_file was extracted to
    """
    try:
        if os.path.isfile(zipped_file):
            # create a ZipFile object and load zipped_file in it
            with zipfile.ZipFile(zipped_file, 'r') as zip_obj:
                # extract all the contents of zipped_file in unzip_directory
                zip_obj.extractall(unzip_dir)
    except zipfile.BadZipFile:
        print('WARNING Given file ' + zipped_file + ' is not a zip file, it cannot be extracted.')
        return zipped_file
    return unzip_dir


def get_filenames_in_dir(directory, including_zip=False):
    """
        Get name(s) of all files in a given directory dir except for those beginning with '.' and/or ending with '.zip'.
        This method is used for getting names of extracted file names.

        :param directory: path to an existing directory
        :param including_zip: True if .zip files are supposed be returned as well
        :return: name(s) of files
    """
    if not os.path.isdir(directory):
        return ''

    names = []

    for dir_path, dir_names, filenames in os.walk(directory):
        for file in filenames:
            if (not including_zip and '.zip' in file) or file.startswith('.'):
                continue
            else:
                names.append(os.path.join(dir_path, file))

    # if only one file was extracted, get its name immediately (that is the case for most analytical modules):
    if len(names) == 1:
        return names[0]
    return names

