#! /usr/bin/env python3
# -*- coding: utf-8 -*-


"""
Script:
1. Loads indexed data from REST API.
2. Extracts them (they are in a zipped folder).
3. Uploads them to database.

Usage:
Independent use:
* Obtain data via REST API:
    $ communication_handler.py -im -in <input_dir> -ou <output_dir> -i <input_identifier> -ip <API address:port>
    --lru_mb <lru_mb> --zero <ip:port> -my=<ip:port> -ap 0/p/ -aa <aplha_args>

* Obtain data via local directory:
    $ communication_handler.py -im -l -in <input_dir> -ou <output_dir> -i <input_identifier> -ip <API address:port>
    --lru_mb <lru_mb> --zero <ip:port> -my=<ip:port> -ap 0/p/ -aa <aplha_args>

Integrated to Orchestrator:
    $ ./communication_handler
        where environment variables PARAMETER_<PREFIX>_LOCAL, DATAWAREHOUSE, PARAMETER_<PREFIX>_INPUT_DIR, 
        PARAMETER_<PREFIX>_OUTPUT_DIR, PARAMETER_<PREFIX>_INPUT_IDENTIFIER, PARAMETER_<PREFIX>_LRU_MB, 
        ENDPOINTS_HANDLING_ZERO_GRPC, MODULE_ID_SELF, PARAMETER_<PREFIX>_ALPHA_P,
        PARAMETER_<PREFIX>_ALPHA_O and PARAMETER_<PREFIX>_ALPHA_ARGS have to be set
"""

import requests
import sys
import re
import fnmatch
import os
import argparse
import ipaddress
import subprocess
import multiprocessing
from pathlib import Path
sys.path.insert(0, str(Path( Path(__file__).parent / Path('python-analysis-modules') ).resolve()) )
import python_analysis_modules.warehouse_connector as warehouse_connector
import python_analysis_modules.zip_functions as zip_functions


def is_dir_path(string):
    """
        Check if the given string is a path to an existing directory.

        :param string: String that should be verified if it is a directory
        :return: True if given string is a path to an existing directory
    """
    return os.path.isdir(string)


def is_abs_dir_path(string):
    """
        Checks if string is an absolute path to an existing directory.

        :return: True if given string is an absolute path to an existing directory
    """
    return len(string) > 0 and string[0] == '/' and is_dir_path(string)


def is_ip_address(string):
    """
        Checks if string is a valid IP address.

        :return: True if string is a valid IP address
    """
    try:
        ipaddress.ip_address(string)
        return True
    except ValueError:
        return False


def is_hostname(string):
    """
        Checks if string is a valid hostname.
        Source: https://stackoverflow.com/questions/2532053/validate-a-hostname-string

        :return: True if string is a valid hostname
    """
    if len(string) > 255:
        return False
    if string[-1] == '.':
        string = string[:-1]
    allowed = re.compile('(?!-)[A-Z\d-]{1,63}(?<!-)$', re.IGNORECASE)
    return all(allowed.match(x) for x in string.split('.'))


def ip_address_or_hostname(string):
    """
        Checks if string is a valid ip_address or hostname.

        :return: string if True
    """
    return is_ip_address(string) or is_hostname(string)


def is_port(val):
    """
        Check if the given val is a valid port number.

        :param val: port number expected to be in range [0, 65536]
        :return: True if val is in valid format
    """
    return is_integer(val) and 0 < val < 65536


def is_ip_and_port(string):
    """
        Check if the given string is in valid format of 'ip:port' or 'hostname:port'.

        :param string: string that contains a : (e.g. '121.11.10.9:1234')
        :return: True if the string is in valid format
    """
    parts = string.split(":")

    if len(parts) != 2:
        return False

    return ip_address_or_hostname(parts[0]) or is_port(parts[1])


def is_boolean(val):
    """
        Check if the given value is a bool.

        :param val: input value
        :return: True if it is an bool
    """
    try:
        eval(val)
        return True
    except NameError:
        # conversion to bool was not successful
        return False


def is_string(val):
    """
        Checks if input val is a string.

        :return: True if val is a string
    """
    return isinstance(val, str)


def is_int(val):
    """
        Checks if input val is an integer.

        :return: True if val is a integer
    """
    try:
        int(val)
        return True
    except ValueError:
        return False


def is_alpha_p(val):
    """
        Checks if input val is a prefix string.

        :return: True if val is a prefix
    """
    return is_string(val) and is_int(val.split('/')[0])


def is_prefix(val):
    """
        Checks if input val is a prefix string.

        :return: True if val is a prefix
    """
    return is_string(val) and '-' in val


def check_file_availability(input_dir, file_name):
    file_dir = input_dir + '/' + file_name
    if os.path.isfile(file_dir):
        return file_dir
    else:
        print('ERROR file ' + file_dir + ' not found')
        return ''


def check_input_name_substr(input_name, substring):
    return substring in os.path.basename(input_name)


def get_subdir_path(input_zip, unzip_dir, subdir):
    extracted = zip_functions.extract_file(input_zip, unzip_dir + '/' + subdir)
    subdirectories = get_immediate_subdirectories(extracted)

    if len(subdirectories) == 1:
        subdir = subdirectories[0]
    else:
        print('ERROR zip file ' + input_zip + ' does not contain indexed files in /p directory.')
        return ''

    extracted_subdir = extracted + '/' + subdir
    return extracted_subdir


def get_input(local, input_dir, input_identifier, output_dir, alpha_p, warehouse_connector):
    # if "0/p/" given as alpha_p, check if any of the input files contains "0" (it can have prefix -> check substr)
    alpha_p = alpha_p.split('/')[0]
    if local:
        input_zip = check_file_availability(input_dir, input_identifier)
    else:
        downloaded_file = warehouse_connector.download_input()
        input_zip = ''
        if downloaded_file:
            # extract downloaded_file to output_dir (skip if its not a zip file):
            extracted_dir_path = zip_functions.extract_file(downloaded_file, output_dir)
            extracted_files = zip_functions.get_filenames_in_dir(extracted_dir_path, True)

            # more than one files were extracted (REDUCE_SHARD in Indexing Module was > 1), I need to find only the
            # one that I'm interested in (specified by argument ALPHA_P value)
            if extracted_files and isinstance(extracted_files, list):
                for file in extracted_files:
                    if check_input_name_substr(file, alpha_p):
                        input_zip = file
                        break
            else:
                input_zip = extracted_files

    print('Zip file ' + input_zip + ' will be used to search for other zipped files.')
    return get_subdir_path(input_zip, output_dir, alpha_p)


def get_immediate_subdirectories(dir_path):
    return [name for name in os.listdir(dir_path) if os.path.isdir(os.path.join(dir_path, name))]


def get_prefix():
    """
        Called if independent_mode is not set. Gets name of module in Orchestrator from MODULE_ID_SELF and transforms
        it to prefix that is used as prefix for other environment variables.
    """
    print('Extracting prefix for environment variables from module name:\n')
    module_name = get_arg_value('MODULE_NAME', '', False, is_string)
    formatted_prefix = 'PARAMETER_' + module_name.upper().replace('-', '_') + '_'
    print('\nPrefix successfully extracted as "' + formatted_prefix + '"\n')
    return formatted_prefix


def get_arg_value(name, prefix, use_prefix, check_function):
    """
        Tries to get value of environment variable name. If the env var is defined, it checks it its value is
        in correct format. If some required variable could not be found, the script will call sys.exit(1).

        :param name: name of environment variable
        :param prefix: environment variables prefix (corresponds to the module name in Orchestrator db)
        :param use_prefix: True = prefix will be concatenated to name
        :param check_function: name of a function that checks if the value of env var is in correct format
    """
    name = prefix + name if use_prefix else name
    if name in os.environ:
        value = os.getenv(name)
        if check_function(value):
            print('OK      {:<32} {}'.format(name, value))
            return value
        else:
            print('ERROR   {:<32} {} - is not in correct format'.format(name, value))
    else:
        print('ERROR   {:<32} is not defined or could not be read'.format(name))
        # if this is printed check module definition on orchestrator database, there
        # should always be at least some default value returned

    print('\nExiting.')
    sys.exit(1)
    # something went wrong (either the environment variable value could not be read or it was in incorrect format)
    # therefore we do not want to continue with execution of this script - we consider each variable to be required (if
    # both of the conditions are not satisfied, it indicates some problem in communication with the database etc.)


def define_env_arguments(prefix):
    """
        Loads all environment variables to a dictionary { env_var_name : env_val_value } and checks
        whether their values are in expected format. If any value is either missing or in incorrect format the
        execution of this script is terminated.
    """
    args_dict = {}

    # arg_names contains similar information to argparse.ArgumentParser() for each env var
    arg_names = [('LOCAL', is_boolean, True),
                 ('DATAWAREHOUSE', is_ip_and_port, False),
                 ('INPUT_DIR', is_dir_path, True),
                 ('OUTPUT_DIR', is_dir_path, True),
                 ('INPUT_IDENTIFIER', is_string, True),

                 # dgraph alpha command arguments:
                 ('LRU_MB', is_int, True),
                 ('ENDPOINT_HANDLING_ZERO_GRPC', is_ip_and_port, False),
                 ('MODULE_ID_SELF', ip_address_or_hostname, False),
                 ('ALPHA_P', is_alpha_p, True),
                 ('ALPHA_O', is_string, True),
                 ('CUSTOM_TOKENIZERS', is_string, True),
                 ('ALPHA_ARGS', is_string, True)]

    print('Loading environment variables:\n\nStatus  Name                             Value')
    for name, check_function, use_prefix in arg_names:
        value = get_arg_value(name, prefix, use_prefix, check_function)
        if value:
            args_dict[name] = value
        else:
            args_dict[name] = ''
    print('\nSuccessfully loaded environment variables.\n')

    return args_dict


def define_parsed_arguments():
    """
        Add arguments to ArgumentParser (argparse) module instance.

        Argument -m (--independent_mode) is a switch that indicates that this script will run in an independent mode
        (not integrated in the Orchestrator). The arguments are then expected to be passed as command line arguments,
        not environment variables.

        :return: Parsed arguments
    """
    parser = argparse.ArgumentParser()

    parser.add_argument('-im', '--independent_mode', help='Switch to indicate that this script is running in an '
                        'independent mode (not integrated in the Orchestrator) - the arguments are then expected to be '
                        'passed as command line arguments, not environment variables.', action='store_true')

    parser.add_argument('-in', '--input_dir', help='Input directory', default='/usr/local/bin/granef/data/input/')
    parser.add_argument('-ou', '--output_dir', help='Output directory', default='/usr/local/bin/granef/data/output/')
    parser.add_argument('-i', '--input_identifier', help='Name/uuid of a zipped file that contains the data',
                        default='', required='-im' in sys.argv or '--independent_mode' in sys.argv)
    parser.add_argument('-ip', '--ip_and_port', help='IP address and port of REST API', default='127.0.0.1:5000')
    parser.add_argument('-l', '--local', help='Use a local directory instead of REST API', action='store_true')

    # dgraph alpha command arguments
    parser.add_argument('-lr', '--lru_mb', help='Alpha command "--lru_mb" argument value', default='2048')
    parser.add_argument('-z', "--zero", help='Alpha command "--zero" argument value', default='localhost:5080')
    parser.add_argument('-my', "--my", help='Alpha command "--my" argument value', default='localhost:7080')
    parser.add_argument('-ap', "--alpha_p", help='Alpha command "-p" argument value (data directory specifier)',
                        default='0/p/')
    parser.add_argument('-o', '--alpha_o', help='Alpha command "-o" argument value (port offset for another Alpha)',
                        default="")
    parser.add_argument('-c', '--custom_tokenizers', help='Bulk command "--custom_tokenizers" argument value',
                        default='cidr-plugin.so')
    parser.add_argument('-aa', '--alpha_args', help='Other alpha command arguments but in format "--arg value"',
                        default='')

    return parser.parse_args()


def check_and_load_parsed_arg(args_dict, script_name, arguments, arg_name, check_func, msg):
    """
        Check if argument with name arg_name from argparse module instance arguments is in the correct format according
        to check_func. If yes - then store its value in dictionary args_dict under name script_name. If not - print out
        msg and raise ArgumentTypeError.

        :param args_dict: arguments dictionary used later in main
        :param script_name: name of an argument used in this script
        :param arguments: argparse module instance with command line argument values
        :param arg_name: name of concrete argument from argparse
        :param check_func: what function should is used for value check of argument arg_name
        :param msg: what msg will be printed out if the value of argument is incorrect
    """
    value = getattr(arguments, arg_name)
    if check_func(value):
        args_dict[script_name] = value
    else:
        raise argparse.ArgumentTypeError(msg + value)


def resolve_parsed_names(arguments):
    """
        Helper function to make the use of environment variables and command line arguments easier in main.

        :param arguments: argparse module instance with command line argument values
        :return: Dictionary with converted command line arguments names and their values
    """
    args_dict = {}
    args_dict['LOCAL'] = str(getattr(arguments, 'local'))
    check_and_load_parsed_arg(args_dict, 'DATAWAREHOUSE', arguments, 'ip_and_port', is_ip_and_port,
                              'given string is not a valid <ip:port>: ')
    check_and_load_parsed_arg(args_dict, 'INPUT_DIR', arguments, 'input_dir', is_dir_path,
                              'given string is not a valid path to an existing directory: ')
    check_and_load_parsed_arg(args_dict, 'OUTPUT_DIR', arguments, 'output_dir', is_dir_path,
                              'given string is not a valid path to an existing directory: ')
    check_and_load_parsed_arg(args_dict, 'INPUT_IDENTIFIER', arguments, 'input_identifier', is_string,
                              'given string is not a valid ZIP file name: ')

    check_and_load_parsed_arg(args_dict, 'LRU_MB', arguments, 'lru_mb', is_int,
                              'given string is not a valid lru_mb: ')
    check_and_load_parsed_arg(args_dict, 'ENDPOINT_HANDLING_ZERO_GRPC', arguments, 'zero', is_ip_and_port,
                              'given string is not a valid <ip:port>: ')
    check_and_load_parsed_arg(args_dict, 'MODULE_ID_SELF', arguments, 'my', is_ip_and_port,
                              'given string is not a valid ip or hostname: ')
    check_and_load_parsed_arg(args_dict, 'ALPHA_P', arguments, 'alpha_p', is_alpha_p,
                              'given string is not a valid alpha_p: ')
    check_and_load_parsed_arg(args_dict, 'ALPHA_O', arguments, 'alpha_o', is_string,
                              'given string is not a valid alpha_o: ')
    check_and_load_parsed_arg(args_dict, 'CUSTOM_TOKENIZERS', arguments, 'custom_tokenizers', is_string,
                              'given string is not a valid custom_tokenizers string: ')
    check_and_load_parsed_arg(args_dict, 'ALPHA_ARGS', arguments, 'alpha_args', is_string,
                              'given string is not a valid alpha_args: ')
    return args_dict


if __name__ == '__main__':
    args_loaded = define_parsed_arguments()

    if getattr(args_loaded, 'independent_mode'):
        args = resolve_parsed_names(args_loaded)
    else:
        prefix = get_prefix()
        args = define_env_arguments(prefix)
        # this environment variables prefix corresponds to the module name in orchestrator db

    address = 'http://' + args['DATAWAREHOUSE'] + '/'

    input_dir = args['INPUT_DIR']
    output_dir = args['OUTPUT_DIR']
    input_identifier = args['INPUT_IDENTIFIER']
    alpha_p = args['ALPHA_P']

    wc = warehouse_connector.WarehouseConnector(input_identifier,
                                                input_dir,
                                                '',
                                                '',
                                                False,
                                                '',
                                                'dgraph_indexing',
                                                None,
                                                None,
                                                address)

    data_input = get_input(eval(args['LOCAL']), input_dir, input_identifier, output_dir, alpha_p, wc)

    if data_input and os.path.isdir(data_input):
        other_args = ''
        arg_o = args['ALPHA_O']
        arg_other = args['ALPHA_ARGS']
        custom_tokenizers = args['CUSTOM_TOKENIZERS']
        if arg_o:
            other_args += '-o=' + arg_o
        if custom_tokenizers:
            other_args += ' --custom_tokenizers=' + custom_tokenizers.strip()
        if arg_other:
            other_args += ' ' + arg_other

        my = args['MODULE_ID_SELF']
        if ':' not in my:
            my += ':7080'
            # for usage in Orchestrator, port 7080 will be used always, in independent_mode a different port can be used

        # running Dgraph Alpha uploads the data (from directory output_dir + getattr(args, 'alpha_p')) to graph db
        try:
            print(' dgraph alpha --lru_mb ' + args['LRU_MB'] + ' --zero ' + args['ENDPOINT_HANDLING_ZERO_GRPC']
                  + ' --my ' + my + ' -p ' + data_input + ' ' + other_args)

            subprocess.run(' dgraph alpha --lru_mb ' + args['LRU_MB'] + ' --zero ' + args['ENDPOINT_HANDLING_ZERO_GRPC']
                           + ' --my ' + my + ' -p ' + data_input + ' ' + other_args, shell=True, check=True)

        except subprocess.CalledProcessError as e:
            print('Error (CalledProcessError) when running dgraph alpha command:\nCommand =' + str(e.cmd) +
                  ',\nReturncode =' + str(e.returncode) + ',\nOutput =' + str(e.output))
        except Exception as e:
            print('Error when running dgraph alpha command (' + str(e) + ').')

    else:
        print('Input zipped data couldn\'t be downloaded, Dgraph Alpha didn\'t start.')
