#! /usr/bin/env python3
# -*- coding: utf-8 -*-


"""
Script:
For handling Dgraph zero command arguments.

Usage:  $ ./zero_arg_handler.py im -my <ip:port> --bindall <boolean>

Integrated to Orchestrator:
        $ python3 communication_handler.py
            where environment variables MODULE_ID_SELF and PARAMETER_<PREFIX>_BINDALL have to be set
"""

import sys
import os
import re
import argparse
import ipaddress
import subprocess


def is_string(val):
    """
        Checks if input val is a string.

        :return: True if val is a string
    """
    return isinstance(val, str)


def is_ip_address(string):
    """
        Checks if string is a valid IP address.

        :return: True if string is a valid IP address
    """
    try:
        ipaddress.ip_address(string)
        return True
    except ValueError:
        return False


def is_hostname(string):
    """
        Checks if string is a valid hostname.
        Source: https://stackoverflow.com/questions/2532053/validate-a-hostname-string

        :return: True if string is a valid hostname
    """
    if len(string) > 255:
        return False
    if string[-1] == '.':
        string = string[:-1]
    allowed = re.compile('(?!-)[A-Z\d-]{1,63}(?<!-)$', re.IGNORECASE)
    return all(allowed.match(x) for x in string.split('.'))


def is_ip_address_or_hostname(string):
    """
        Checks if string is a valid ip_address or hostname.

        :return: string if True
    """
    return is_ip_address(string) or is_hostname(string)


def is_port(val):
    """
        Check if the given val is a valid port number.

        :param val: port number expected to be in range [0, 65536]
        :return: True if val is in valid format
    """
    return is_integer(val) and 0 < val < 65536


def is_ip_and_port(string):
    """
        Check if the given string is in valid format of 'ip:port' or 'hostname:port'.

        :param string: string that contains a : (e.g. '121.11.10.9:1234')
        :return: True if the string is in valid format
    """
    parts = string.split(":")

    if len(parts) != 2:
        return False

    return is_ip_address_or_hostname(parts[0]) or is_port(parts[1])


def is_boolean(val):
    """
        Check if the given value is a bool.

        :param val: input value
        :return: True if it is an bool
    """
    try:
        eval(val)
        return True
    except NameError:
        # conversion to bool was not successful
        return False


def is_prefix(val):
    """
        Checks if input val is a prefix string.

        :return: True if val is a prefix
    """
    return isinstance(val, str) and '-' in val


def get_prefix():
    """
        Called if independent_mode is not set. Gets name of module in Orchestrator from MODULE_ID_SELF and transforms
        it to prefix that is used as prefix for other environment variables.
    """
    print('Extracting prefix for environment variables from module name:\n')
    module_name = get_arg_value('MODULE_NAME', '', False, is_string)
    formatted_prefix = 'PARAMETER_' + module_name.upper().replace('-', '_') + '_'
    print('\nPrefix successfully extracted as "' + formatted_prefix + '"\n')
    return formatted_prefix


def get_arg_value(name, prefix, use_prefix, check_function):
    """
        Tries to get value of environment variable name. If the env var is defined, it checks it its value is
        in correct format. If some required variable could not be found, the script will call sys.exit(1).

        :param name: name of environment variable
        :param check_function: name of a function that checks if the value of env var is in correct format
    """
    name = prefix + name if use_prefix else name
    if name in os.environ:
        value = os.getenv(name)
        if check_function(value):
            print('OK      {:<32} {}'.format(name, value))
            return value
        else:
            print('ERROR   {:<32} {} - is not in correct format'.format(name, value))
    else:
        print('ERROR   {:<32} is not defined or could not be read'.format(name))
        # if this is printed check module definition on orchestrator database, there
        # should always be at least some default value returned

    print('\nExiting.')
    sys.exit(1)
    # something went wrong (either the environment variable value could not be read or it was in incorrect format)
    # therefore we do not want to continue with execution of this script - we consider each variable to be required (if
    # both of the conditions are not satisfied, it indicates some problem in communication with the database etc.)


def define_env_arguments(prefix):
    """
        Loads all environment variables to a dictionary { env_var_name : env_val_value } and checks
        whether their values are in expected format. If any value is either missing or in incorrect format the
        execution of this script is terminated.
    """
    args_dict = {}

    # arg_names contains similar information to argparse.ArgumentParser() for each env var
    arg_names = [('MODULE_ID_SELF', is_ip_address_or_hostname, False),
                 ('BINDALL', is_boolean, True)]

    print('Loading environment variables:\n\nStatus  Name                    Value')
    for name, check_function, use_prefix in arg_names:
        value = get_arg_value(name, prefix, use_prefix, check_function)
        if value:
            args_dict[name] = value
        else:
            args_dict[name] = ''
    print('\nSuccessfully loaded environment variables.\n')

    return args_dict


def define_parsed_arguments():
    """
        Add arguments to ArgumentParser (argparse) module instance.

        Argument -m (--independent_mode) is a switch that indicates that this script will run in an independent mode
        (not integrated in the Orchestrator). The arguments are then expected to be passed as command line arguments,
        not environment variables.

        :return: Parsed arguments
    """
    parser = argparse.ArgumentParser()

    parser.add_argument('-im', '--independent_mode', help='Switch to indicate that this script is running in an '
                        'independent mode (not integrated in the Orchestrator) - the arguments are then expected to be '
                        'passed as command line arguments, not environment variables.', action='store_true')

    parser.add_argument('-my', "--my", help='Zero command "--my" argument value', default='localhost:5080')
    parser.add_argument('-ba', '--bindall', help='Zero command "--bindall" argument value', default="True")

    return parser.parse_args()


def check_and_load_parsed_arg(args_dict, script_name, arguments, arg_name, check_func, msg):
    """
        Check if argument with name arg_name from argparse module instance arguments is in the correct format according
        to check_func. If yes - then store its value in dictionary args_dict under name script_name. If not - print out
        msg and raise ArgumentTypeError.

        :param args_dict: arguments dictionary used later in main
        :param script_name: name of an argument used in this script
        :param arguments: argparse module instance with command line argument values
        :param arg_name: name of concrete argument from argparse
        :param check_func: what function should is used for value check of argument arg_name
        :param msg: what msg will be printed out if the value of argument is incorrect
    """
    value = getattr(arguments, arg_name)
    if check_func(value):
        args_dict[script_name] = value
    else:
        raise argparse.ArgumentTypeError(msg + value)


def resolve_parsed_names(arguments):
    """
        Helper function to make the use of environment variables and command line arguments easier in main.

        :param arguments: argparse module instance with command line argument values
        :return: Dictionary with converted command line arguments names and their values
    """
    args_dict = {}
    check_and_load_parsed_arg(args_dict, 'MODULE_ID_SELF', arguments, 'my', is_ip_and_port,
                              'given string is not a valid <ip:port>: ')
    check_and_load_parsed_arg(args_dict, 'BINDALL', arguments, 'bindall', is_boolean,
                              'given string is not a valid boolean: ')
    return args_dict


if __name__ == '__main__':
    args_loaded = define_parsed_arguments()

    if getattr(args_loaded, 'independent_mode'):
        args = resolve_parsed_names(args_loaded)
    else:
        prefix = get_prefix()
        args = define_env_arguments(prefix)

    my = args['MODULE_ID_SELF']
    bindall = args['BINDALL']

    if ':' not in my:
        my += ':5080'
        # for usage in Orchestrator, port 5080 will be used always,
        # in independent_mode a different port can be used

    try:
        subprocess.run(' dgraph zero --my=' + my + ' --bindall=' + bindall, shell=True, check=True)

    except subprocess.CalledProcessError as e:
        print('Error (CalledProcessError) when running dgraph zero command:\nCommand =' + str(e.cmd) +
              ',\nReturncode =' + str(e.returncode) + ',\nOutput =' + str(e.output))
    except Exception as e:
        print('Error when running dgraph zero command (' + str(e) + ').')

