                                                                                                                #! /usr/bin/env python3
# -*- coding: utf-8 -*-


"""
Script:
1. Loads RDF and schema files from REST API.
2. Applies indexing (Dgraph Zero).
3. Uploads the result back to REST API.

Usage:=
Independent use:
* Obtain data via REST API:
    $ communication_handler.py -im -i <input_identifier> -ip <API address:port> -in <input_dir> -ou <output_dir> 
      -zd <zip_dir_path> -zp <zip_folder_prefix> -ms <map_shards> -rs <reduce_shards> -z <zero>

* Obtain data via local directory:
    $ communication_handler.py -im -l -i <input_identifier> -in <input_dir> -ou <output_dir> -zd <zip_dir_path> 
      -zp <zip_folder_prefix> -ms <map_shards> -rs <reduce_shards> -z <zero>

Integrated to Orchestrator:
    $ ./communication_handler
    where environment variables PARAMETER_<PREFIX>_LOCAL, DATAWAREHOUSE, PARAMETER_<PREFIX>_INPUT_DIR, 
    PARAMETER_<PREFIX>_OUTPUT_DIR, PARAMETER_<PREFIX>_ZIP_FOLDER_PATH, PARAMETER_<PREFIX>_ZIP_FOLDER_PREFIX, 
    PARAMETER_<PREFIX>_INPUT_IDENTIFIER, PARAMETER_<PREFIX>_MAP_SHARDS, PARAMETER_<PREFIX>_REDUCE_SHARDS, 
    PARAMETER_<PREFIX>_ZERO and PARAMETER_<PREFIX>_BULK_ARGS have to be set
"""

import requests
import sys
import re
import fnmatch
import os
import argparse
import ipaddress
import subprocess
import multiprocessing
from pathlib import Path
sys.path.insert(0, str(Path( Path(__file__).parent / Path('python-analysis-modules') ).resolve()) )
import python_analysis_modules.warehouse_connector as warehouse_connector
import python_analysis_modules.zip_functions as zip_functions


def is_string(val):
    """
        Checks if input val is a string.

        :return: True if val is a string
    """
    return isinstance(val, str)


def is_int(val):
    """
        Checks if input val is an integer.

        :return: True if val is a integer
    """
    try:
        int(val)
        return True
    except ValueError:
        return False


def is_boolean(val):
    """
        Check if the given value is a bool.

        :param val: input value
        :return: True if it is an bool
    """
    try:
        eval(val)
        return True
    except NameError:
        # conversion to bool was not successful
        return False


def is_dir_path(string):
    """
        Check if the given string is a path to an existing directory.

        :param string: String that should be verified if it is a directory
        :return: True if given string is a path to an existing directory
    """
    return os.path.isdir(string)


def is_abs_dir_path(string):
    """
        Checks if string is an absolute path to an existing directory.

        :return: True if given string is an absolute path to an existing directory
    """
    return len(string) > 0 and string[0] == '/' and is_dir_path(string)


def is_rdf_file(string):
    """
        Check if the given val is a in RDF file name format.

        :param string: String that should be verified if it is in a correct format
        :return: True if given val is in correct format
    """
    return fnmatch.fnmatch(string, '*.rdf')


def is_schema_file(string):
    """
        Check if the given val is a in SCHEMA file name format.

        :param string: String that should be verified if it is in a correct format
        :return: True if given val is in correct format
    """
    return fnmatch.fnmatch(string, '*.schema')


def is_ip_address(string):
    """
        Checks if string is a valid IP address.

        :return: True if string is a valid IP address
    """
    try:
        ipaddress.ip_address(string)
        return True
    except ValueError:
        return False


def is_hostname(string):
    """
        Checks if string is a valid hostname.
        Source: https://stackoverflow.com/questions/2532053/validate-a-hostname-string

        :return: True if string is a valid hostname
    """
    if len(string) > 255:
        return False
    if string[-1] == '.':
        string = string[:-1]
    allowed = re.compile('(?!-)[A-Z\d-]{1,63}(?<!-)$', re.IGNORECASE)
    return all(allowed.match(x) for x in string.split('.'))


def ip_address_or_hostname(string):
    """
        Checks if string is a valid ip_address or hostname.

        :return: string if True
    """
    return is_ip_address(string) or is_hostname(string)


def is_ip_and_port(string):
    """
        Check if the given string is in valid format of 'ip:port' or 'hostname:port'.

        :param string: string that contains a : (e.g. '121.11.10.9:1234')
        :return: True if the string is in valid format
    """
    parts = string.split(':')

    if len(parts) != 2:
        return False

    return ip_address_or_hostname(parts[0]) or is_integer(parts[1])


def is_prefix(val):
    """
        Checks if input val is a prefix string.

        :return: True if val is a prefix
    """
    return is_string(val) and '-' in val


def get_input_paths(input_dir):
    """
        Check file rdf, schema files availability in input_dir directory.

        :param input_dir: path to input directory
        :return: list of local input file paths
    """
    rdf_path = ''
    schema_path = ''

    for folder_name, _, filenames in os.walk(input_dir):
        for file in filenames:
            if is_rdf_file(file):
                rdf_path = str(os.path.join(folder_name, file))
            elif is_schema_file(file):
                schema_path = str((os.path.join(folder_name, file)))

            # if the files were found I can stop searching for them
            if rdf_path and schema_path:
                break

    return rdf_path, schema_path


def check_input_availability(dir_path, input_name):
    """
        Check file availability in input_dir directory if the "local" argument is set to True.

        :param dir_path: path to input directory
        :param input_name: name of input file
        :return: local input file path
    """
    file_path = dir_path + '/' + input_name
    if os.path.isfile(file_path):
        return file_path
    else:
        print('ERROR file ' + file_path + ' not found')
        return ''


def get_inputs(local, dir_path, input_identifier, warehouse_connector):
    """
        Check availability/ download input file. Extract it and check if it contains rdf and schema files.

        :param local: True is local storage is supposed to be used instead of Data Warehouse
        :param dir_path: path to the directory where the input should be located/ downloaded to
        :param input_identifier: name/ uuid of input file
        :param warehouse_connector: WarehouseConnector instance
        :return: list (path to rdf file, path to schema file)
    """
    if not local:
        downloaded_file = warehouse_connector.download_input()
    else:
        downloaded_file = check_input_availability(dir_path, input_identifier)

    if downloaded_file:
        # extract downloaded_file to input_dir (skip if its not a zip file):
        zip_functions.extract_file(downloaded_file, dir_path)

    return get_input_paths(dir_path)


def get_immediate_subdirectories(output_dir):
    return [name for name in os.listdir(output_dir) if os.path.isdir(os.path.join(output_dir, name))]


def zip_output(dir_path, zip_dir_path, zip_prefix):
    """
        Zips generated directory with dir_path to zip_dir_path.

        :param dir_path: path to the directory that is supposed to be zipped
        :param zip_dir_path: absolute path where the result will be saved
        :param zip_prefix: string that will be concatenated as prefix to name of the final zip file
    """
    to_replace = []
    # zip each directory in output_dir directory
    generated_dirs = get_immediate_subdirectories(dir_path)
    for dir_name in generated_dirs:
        dir_zip_name = zip_prefix + dir_name + '.zip'
        to_zip_dir_name = dir_path + '/' + dir_name
        if zip_functions.zip_files_in_dir(to_zip_dir_name, zip_dir_path, dir_zip_name, ['']):
            to_replace.append((zip_dir + '/' + dir_zip_name, dir_path + '/' + dir_zip_name))
            print('Directory ' + to_zip_dir_name + ' zipped to file ' + dir_zip_name)

    for old, new in to_replace:
        os.replace(old, new)
        print('Replaced ' + old + ' to ' + new)


def upload_zipped_output(output_dir, zip_dir, warehouse_connector):
    """
        Zips generated output files and uploads them to the Data Warehouse using warehouse_connector.

        :param output_dir: path to directory containing output output files
        :param zip_dir: absolute path to a directory where the final output zip file will be saved
        :param warehouse_connector: WarehouseConnector instance
    """
    # zip all output log files to a single output file:
    zip_name = warehouse_connector.get_output_name()
    if zip_functions.zip_files_in_dir(output_dir, zip_dir, zip_name, ['.zip']):
        warehouse_connector.upload_output()
    else:
        print('ERROR file (' + zip_name + ') containing the output files could not be zipped.')


def get_prefix():
    """
        Called if independent_mode is not set. Gets name of module in Orchestrator from MODULE_ID_SELF and transforms
        it to prefix that is used as prefix for other environment variables.
    """
    print('Extracting prefix for environment variables from module name:\n')
    module_name = get_arg_value('MODULE_NAME', '', False, is_string)
    formatted_prefix = 'PARAMETER_' + module_name.upper().replace('-', '_') + '_'
    print('\nPrefix successfully extracted as "' + formatted_prefix + '"\n')
    return formatted_prefix


def get_arg_value(name, prefix, use_prefix, check_function):
    """
        Tries to get value of environment variable name. If the env var is defined, it checks it its value is
        in correct format. If some required variable could not be found, the script will call sys.exit(1).

        :param name: name of environment variable
        :param prefix: environment variables prefix (corresponds to the module name in Orchestrator db)
        :param use_prefix: True = prefix will be concatenated to name
        :param check_function: name of a function that checks if the value of env var is in correct format
    """
    name = prefix + name if use_prefix else name
    if name in os.environ:
        value = os.getenv(name)
        if check_function(value):
            print('OK      {:<42} {}'.format(name, value))
            return value
        else:
            print('ERROR   {:<42} {} - is not in correct format'.format(name, value))
    else:
        print('ERROR   {:<42} is not defined or could not be read'.format(name))
        # if this is printed check module definition on orchestrator database, there
        # should always be at least some default value returned

    print('\nExiting.')
    sys.exit(1)
    # something went wrong (either the environment variable value could not be read or it was in incorrect format)
    # therefore we do not want to continue with execution of this script - we consider each variable to be required (if
    # both of the conditions are not satisfied, it indicates some problem in communication with the database etc.)


def define_env_arguments(prefix):
    """
        Loads all environment variables to a dictionary { env_var_name : env_val_value } and checks
        whether their values are in expected format. If any value is either missing or in incorrect format the
        execution of this script is terminated.
    """
    args_dict = {}

    arg_names = [('LOCAL', is_boolean, True),
                 ('DATAWAREHOUSE', is_ip_and_port, False),
                 ('INPUT_DIR', is_dir_path, True),
                 ('OUTPUT_DIR', is_dir_path, True),
                 ('ZIP_DIR_PATH', is_abs_dir_path, True),
                 ('ZIP_DIR_PREFIX', is_string, True),

                 # dgraph bulk command arguments:
                 ('INPUT_IDENTIFIER', is_string, True),
                 ('MAP_SHARDS', is_int, True),
                 ('REDUCE_SHARDS', is_int, True),
                 ('ZERO', is_ip_and_port, True),
                 ('CUSTOM_TOKENIZERS', is_string, True),
                 ('BULK_ARGS', is_string, True)]

    print('Loading environment variables:\n\nStatus  Name                                       Value')
    for name, check_function, use_prefix in arg_names:
        value = get_arg_value(name, prefix, use_prefix, check_function)
        if value:
            args_dict[name] = value
        else:
            args_dict[name] = ''
    print('\nSuccessfully loaded environment variables.\n')

    return args_dict


def define_parsed_arguments():
    """
        Add arguments to ArgumentParser (argparse) module instance.

        Argument -m (--independent_mode) is a switch that indicates that this script will run in an independent mode
        (not integrated in the Orchestrator). The arguments are then expected to be passed as command line arguments,
        not environment variables.

        :return: Parsed arguments
    """
    parser = argparse.ArgumentParser()

    parser.add_argument('-im', '--independent_mode', help='Switch to indicate that this script is running in an '
                        'independent mode (not integrated in the Orchestrator) - the arguments are then expected to be '
                        'passed as command line arguments, not environment variables.', action='store_true')

    parser.add_argument('-in', '--input_dir', help='Input directory', default='/usr/local/bin/granef/data/input/')
    parser.add_argument('-ou', '--output_dir', help='Output directory', default='/usr/local/bin/granef/data/output/')
    parser.add_argument('-zd', '--zip_dir_path', help='Path to output zip directory',
                        default="/usr/local/bin/granef/data/zip/")
    parser.add_argument('-zp', '--zip_dir_prefix', help='Zip output folder name prefix', default='')
    parser.add_argument('-ip', '--ip_and_port', help='IP address and port of REST API', default='127.0.0.1:5000')
    parser.add_argument('-i', '--input_identifier', help='Input file name/ uuid', default='', required=('-im' in sys.argv
                        or '--independent_mode' in sys.argv) and ('-l' not in sys.argv or '--local' not in sys.argv))

    # dgraph bulk command arguments
    parser.add_argument('-ms', '--map_shards', help='Bulk command "--map_shards" argument value', default='2')
    parser.add_argument('-rs', '--reduce_shards', help='Bulk command "--reduce_shards" argument value', default='1')
    parser.add_argument('-z', '--zero', help='Bulk command "--zero" argument value', default='localhost:5080')
    parser.add_argument('-c', '--custom_tokenizers', help='Bulk command "--custom_tokenizers" argument value',
                        default='cidr-plugin.so')
    parser.add_argument('-b', '--bulk_args', help='Other bulk command arguments but in format "--arg value"',
                        default='')
    parser.add_argument('-l', '--local', help='Use a local directory instead of REST API', action='store_true')

    return parser.parse_args()


def check_and_load_parsed_arg(args_dict, script_name, arguments, arg_name, check_func, msg):
    """
        Check if argument with name arg_name from argparse module instance arguments is in the correct format according
        to check_func. If yes - then store its value in dictionary args_dict under name script_name. If not - print out
        msg and raise ArgumentTypeError.

        :param args_dict: arguments dictionary used later in main
        :param script_name: name of an argument used in this script
        :param arguments: argparse module instance with command line argument values
        :param arg_name: name of concrete argument from argparse
        :param check_func: what function should is used for value check of argument arg_name
        :param msg: what msg will be printed out if the value of argument is incorrect
    """
    value = getattr(arguments, arg_name)
    if check_func(value):
        args_dict[script_name] = value
    else:
        raise argparse.ArgumentTypeError(msg + value)


def resolve_parsed_names(arguments):
    """
        Helper function to make the use of environment variables and command line arguments easier in main.

        :param arguments: argparse module instance with command line argument values
        :return: Dictionary with converted command line arguments names and their values
    """
    args_dict = {}
    args_dict['LOCAL'] = str(getattr(arguments, 'local'))
    check_and_load_parsed_arg(args_dict, 'DATAWAREHOUSE', arguments, 'ip_and_port', is_ip_and_port,
                              'given string is not a valid <ip:port>: ')
    check_and_load_parsed_arg(args_dict, 'INPUT_DIR', arguments, 'input_dir', is_dir_path,
                              'given string is not a valid path to an existing directory: ')
    check_and_load_parsed_arg(args_dict, 'OUTPUT_DIR', arguments, 'output_dir', is_dir_path,
                              'given string is not a valid path to an existing directory: ')
    check_and_load_parsed_arg(args_dict, 'ZIP_DIR_PATH', arguments, 'zip_dir_path', is_abs_dir_path,
                              'given string is not a valid absolute path to an existing directory: ')
    check_and_load_parsed_arg(args_dict, 'ZIP_DIR_PREFIX', arguments, 'zip_dir_prefix', is_string,
                              'given string is not a valid zip_folder_prefix: ')
    check_and_load_parsed_arg(args_dict, 'INPUT_IDENTIFIER', arguments, 'input_identifier', is_string,
                              'given string is not a valid input file name/ uuid: ')

    check_and_load_parsed_arg(args_dict, 'MAP_SHARDS', arguments, 'map_shards', is_int,
                              'given string is not a valid integer: ')
    check_and_load_parsed_arg(args_dict, 'REDUCE_SHARDS', arguments, 'reduce_shards', is_int,
                              'given string is not a valid integer: ')
    check_and_load_parsed_arg(args_dict, 'ZERO', arguments, 'zero', is_ip_and_port,
                              'given string is not a valid <ip:port>: ')
    check_and_load_parsed_arg(args_dict, 'CUSTOM_TOKENIZERS', arguments, 'custom_tokenizers', is_string,
                              'given string is not a valid custom_tokenizers string: ')
    check_and_load_parsed_arg(args_dict, 'BULK_ARGS', arguments, 'bulk_args', is_string,
                              'given string is not a valid bulk_args: ')
    return args_dict


if __name__ == "__main__":
    args_loaded = define_parsed_arguments()

    if getattr(args_loaded, 'independent_mode'):
        args = resolve_parsed_names(args_loaded)
    else:
        prefix = get_prefix()
        args = define_env_arguments(prefix)
        # this environment variables prefix corresponds to the module name in orchestrator db

    address = 'http://' + args["DATAWAREHOUSE"] + '/'
    input_dir = args['INPUT_DIR']
    output_dir = args['OUTPUT_DIR']
    zip_dir = args['ZIP_DIR_PATH']
    local_storage = eval(args['LOCAL'])
    input_identifier = args['INPUT_IDENTIFIER']

    wc = warehouse_connector.WarehouseConnector(input_identifier,
                                                input_dir,
                                                zip_dir,
                                                'indexed',
                                                True,
                                                'dgraph_indexing',
                                                'dgraph_transformation',
                                                ['REDUCE_SHARDS'],
                                                args,
                                                address)
    # check file availability
    input_rdf, input_schema = get_inputs(local_storage, input_dir, input_identifier, wc)

    if input_rdf and input_schema:
        try:
            # run processes in parallel
            zero_process = subprocess.Popen('dgraph zero', shell=True)

            other_args = ''
            reduce_shards = args['REDUCE_SHARDS']
            custom_tokenizers = args['CUSTOM_TOKENIZERS']
            bulk_agrs = args['BULK_ARGS']
            if reduce_shards:
                other_args += '--reduce_shards=' + str(reduce_shards)
            if custom_tokenizers:
                other_args += ' --custom_tokenizers=' + custom_tokenizers.strip()
            if bulk_agrs:
                other_args += ' ' + bulk_agrs

            bulk_process = subprocess.Popen('dgraph bulk -f ' + input_rdf + ' -s ' + input_schema
                                            + ' --map_shards=' + str(args['MAP_SHARDS']) + ' -z ' + args['ZERO']
                                            + ' --replace_out --out ' + output_dir + ' ' + other_args, shell=True)

            # wait for completion
            bulk_process.wait()

        except OSError as e:
            print('ERROR (OSError) when running dgraph (errno=' + str(e.errno) + ', strerror' + str(e.strerror) +
                  ', filename' + str(e.filename) + ').')
        except Exception as e:
            print('ERROR when running dgraph (' + str(e) + ').')

        # zip each directory in output_dir directory and move the final zip files to output_dir
        zip_output(output_dir, zip_dir, args['ZIP_DIR_PREFIX'])

        if not local_storage:
            upload_zipped_output(output_dir, zip_dir, wc)
    else:
        print('Input rdf or schema file couldn\'t be downloaded, nothing was indexed.')
