#! /usr/bin/env python3
# -*- coding: utf-8 -*-


"""
Script:
1. Loads log files (specified as input_identifier) from REST API in JSON format.
2. Those files (if there are any) are then transformed using zeek-dgraph.py.
3. The result is uploaded back to REST API.

Usage:=
Independent use:
* Obtain data via REST API:
    $ communication_handler.py -im -i <input_identifier> -rf <rdf_file_name> -sf <schema_file_name> -c <case_id>
      -ip <API address:port>

* Obtain data via local directory:
    $ communication_handler.py -im -l -i <input_identifier> -rf <rdf_file_name> -sf <schema_file_name> -c <case_id>
      -in <input_dir> -ou <output_dir>

Integrated to Orchestrator:
Usage:  $ ./communication_handler
         where environment variables PARAMETER_<PREFIX>_LOCAL, DATAWAREHOUSE, PARAMETER_<PREFIX>_INPUT_DIR, 
         PARAMETER_<PREFIX>_OUTPUT_DIR, PARAMETER_<PREFIX>_ZIP_DIR, PARAMETER_<PREFIX>_RDF_FILE_NAME, 
         PARAMETER_<PREFIX>_SCHEMA_FILE_NAME, PARAMETER_<PREFIX>_CASE_ID and PARAMETER_<PREFIX>_INPUT_IDENTIFIER 
         have to be set
"""


import requests
import sys
import re
import os
import subprocess
import argparse
import ipaddress
from pathlib import Path
sys.path.insert(0, str(Path( Path(__file__).parent / Path('python-analysis-modules') ).resolve()) )
import python_analysis_modules.warehouse_connector as warehouse_connector
import python_analysis_modules.zip_functions as zip_functions


def is_dir_path(string):
    """
        Checks if string is a path to an existing directory.

        :return: True if given string is a path to an existing directory
    """
    return os.path.isdir(string)


def is_abs_dir_path(string):
    """
        Checks if string is an absolute path to an existing directory.

        :return: True if given string is an absolute path to an existing directory
    """
    return len(string) > 0 and string[0] == '/' and is_dir_path(string)


def is_string(val):
    """
        Checks if input val is a string.

        :return: True if val is a string
    """
    return isinstance(val, str)


def is_ip_address(string):
    """
        Checks if string is a valid IP address.

        :return: True if string is a valid IP address
    """
    try:
        ipaddress.ip_address(string)
        return True
    except ValueError:
        return False


def is_hostname(string):
    """
        Checks if string is a valid hostname.
        Source: https://stackoverflow.com/questions/2532053/validate-a-hostname-string

        :return: True if string is a valid hostname
    """
    if len(string) > 255:
        return False
    if string[-1] == ".":
        string = string[:-1]
    allowed = re.compile("(?!-)[A-Z\d-]{1,63}(?<!-)$", re.IGNORECASE)
    return all(allowed.match(x) for x in string.split("."))


def ip_address_or_hostname(string):
    """
        Checks if string is a valid ip_address or hostname.

        :return: string if True
    """
    return is_ip_address(string) or is_hostname(string)


def is_port(val):
    """
        Check if the given val is a valid port number.

        :param val: port number expected to be in range [0, 65536]
        :return: True if val is in valid format
    """
    return is_integer(val) and 0 < val < 65536


def is_ip_and_port(string):
    """
        Check if the given string is in valid format of 'ip:port' or 'hostname:port'.

        :param string: string that contains API address : (e.g. '121.11.10.9:1234')
        :return: True if the string is in valid format
    """
    parts = string.split(":")

    if len(parts) != 2:
        return False

    return ip_address_or_hostname(parts[0]) or is_port(parts[1])


def is_boolean(val):
    """
        Check if the given value is a bool.

        :param val: input value
        :return: True if it is an bool
    """
    try:
        eval(val)
        return True
    except NameError:
        # conversion to bool was not successful
        return False


def is_prefix(val):
    """
        Checks if input val is a prefix string.

        :return: True if val is a prefix
    """
    return is_string(val) and '-' in val


def get_logs_command(input_dir, file_name):
    """
        Checks if log file is available in input_dir and generates argument containing file_name and path to
        the file. (for example: '--log_conn /usr/local/bin/granef/data/input/conn.log')

        :return: argument for dgraph bulk command
    """
    file_dir = input_dir + file_name + '.log'
    if os.path.isfile(file_dir):
        return ' --log_' + file_name + ' ' + file_dir
    else:
        print('WARNING file ' + file_name + ' not found')
        return ''


def get_input(local, dir_path, file_identifier, log_names, warehouse_connector):
    """
        If local == True checks if input is available, else downloads it from Data Warehouse using warehouse_connector.
        Then extracts the input and generates dgraph bulk command arguments from available files.

        :param local: boolean indicating whether local storage should be used
        :param dir_path: path to a directory where the input is (or will be downloaded to)
        :param file_identifier: input file identifier (uuid)
        :param log_names: list containing log file names
        :param warehouse_connector: WarehouseConnector instance
        :return: script arguments containing all available log files
    """
    if local:
        if file_identifier:
            logs_file = dir_path + '/' + file_identifier
        else:
            logs_file = dir_path
    else:
        logs_file = warehouse_connector.download_input()

    logs_dir = ''
    if logs_file:
        # extract downloaded_file to dir_path (skip if its not a zip file):
        logs_dir = zip_functions.extract_file(logs_file, dir_path)

    arguments = ''
    for file_name in log_names:
        arguments += get_logs_command(logs_dir if logs_dir else dir_path, file_name)

    return arguments


def upload_zipped_output(output_dir, zip_dir, warehouse_connector):
    """
        Zips generated output log files and uploads to the Data Warehouse using warehouse_connector.

        :param output_dir: path to directory containing output log files
        :param zip_dir: absolute path to a directory where the output zip file will be saved
        :param warehouse_connector: WarehouseConnector instance
    """
    # zip all output log files to a single output file:
    zip_name = wc.get_output_name()
    if zip_functions.zip_files_in_dir(output_dir, zip_dir, zip_name, ['.rdf', '.schema']):
        warehouse_connector.upload_output()
    else:
        print('ERROR file (' + zip_name + ') containing the output files could not be zipped.')


def get_prefix():
    """
        Called if independent_mode is not set. Gets name of module in Orchestrator from MODULE_ID_SELF and transforms
        it to prefix that is used as prefix for other environment variables.
    """
    print('Extracting prefix for environment variables from module name:\n')
    module_name = get_arg_value('MODULE_NAME', '', False, is_string)
    formatted_prefix = 'PARAMETER_' + module_name.upper().replace('-', '_') + '_'
    print('\nPrefix successfully extracted as "' + formatted_prefix + '"\n')
    return formatted_prefix


def get_arg_value(name, prefix, use_prefix, check_function):
    """
        Tries to get value of environment variable name. If the env var is defined, it checks it its value is
        in correct format. If some required variable could not be found, the script will call sys.exit(1).

        :param name: name of environment variable
        :param prefix: prefix for environment variables (as specified in Orchestrator db)
        :param use_prefix: boolean indicating that a prefix should be concatenated to name
        :param check_function: name of a function that checks if the value of env var is in correct format
    """
    name = prefix + name if use_prefix else name
    if name in os.environ:
        value = os.getenv(name)
        if check_function(value):
            print('OK      {:<42} {}'.format(name, value))
            return value
        else:
            print('ERROR   {:<42} {} - is not in correct format'.format(name, value))
    else:
        print('ERROR   {:<42} is not defined or could not be read'.format(name))
        # if this is printed check module definition on orchestrator database, there
        # should always be at least some default value returned

    print('\nExiting.')
    sys.exit(1)
    # something went wrong (either the environment variable value could not be read or it was in incorrect format)
    # therefore we do not want to continue with execution of this script - we consider each variable to be required (if
    # both of the conditions are not satisfied, it indicates some problem in communication with the database etc.)


def define_env_arguments(prefix):
    """
        Loads all environment variables to a dictionary { env_var_name : env_val_value } and checks
        whether their values are in expected format. If any value is either missing or in incorrect format the
        execution of this script is terminated.
    """
    args_dict = {}
    arg_names = [('LOCAL', is_boolean, True),
                 ('DATAWAREHOUSE', is_ip_and_port, False),
                 ('INPUT_DIR', is_dir_path, True),
                 ('OUTPUT_DIR', is_dir_path, True),
                 ('ZIP_DIR', is_abs_dir_path, True),
                 ('RDF_FILE_NAME', is_string, True),
                 ('SCHEMA_FILE_NAME', is_string, True),
                 ('CASE_ID', is_string, True),
                 ('INPUT_IDENTIFIER', is_string, True)]

    print('Loading environment variables:\n\nStatus  Name                                       Value')
    for name, check_function, use_prefix in arg_names:
        value = get_arg_value(name, prefix, use_prefix, check_function)
        if value:
            args_dict[name] = value
        else:
            args_dict[name] = ''
    print('\nSuccessfully loaded environment variables.\n')

    return args_dict


def define_parsed_arguments():
    """
        Add arguments to ArgumentParser (argparse) module instance.

        Argument -m (--independent_mode) is a switch that indicates that this script will run in an independent mode
        (not integrated in the Orchestrator). The arguments are then expected to be passed as command line arguments,
        not environment variables.

        :return: Parsed arguments
    """
    parser = argparse.ArgumentParser()

    parser.add_argument('-im', '--independent_mode', help='Switch to indicate that this script is running in an '
                        'independent mode (not integrated in the Orchestrator) - the arguments are then expected to be '
                        'passed as command line arguments, not environment variables.', action='store_true')

    parser.add_argument('-c', '--case_id', type=str, required=('-im' in sys.argv or '--independent_mode' in sys.argv)
                        and ('-l' in sys.argv or '--local' in sys.argv), default='')
    parser.add_argument('-in', '--input_dir', help='Input directory', default='/usr/local/bin/granef/data/input/')
    parser.add_argument('-ou', '--output_dir', help='Output directory', default='/usr/local/bin/granef/data/output/')
    parser.add_argument('-zd', '--zip_dir_path', help='Path to output zip directory',
                        default='/usr/local/bin/granef/data/zip/')
    parser.add_argument('-rf', '--rdf_file_name', help='Output RDF file name', default='zeek-dgraph')
    parser.add_argument('-sf', '--schema_file_name', help='Output schema file name', default='zeek-dgraph')
    parser.add_argument('-i', '--input_identifier', help='Input file uuid containing logs', required=('-im' in sys.argv or
                        '--independent_mode' in sys.argv) and ('-l' not in sys.argv and '--local' not in sys.argv),
                        default='')
    parser.add_argument('-ip', '--ip_and_port', help='IP address and port of REST API', default='127.0.0.1:5000')
    parser.add_argument('-l', '--local', help='Use a local directory instead of REST API', action='store_true')

    return parser.parse_args()


def check_and_load_parsed_arg(args_dict, script_name, arguments, arg_name, check_func, msg):
    """
        Check if argument with name arg_name from argparse module instance arguments is in the correct format according
        to check_func. If yes - then store its value in dictionary args_dict under name script_name. If not - print out
        msg and raise ArgumentTypeError.

        :param args_dict: arguments dictionary used later in main
        :param script_name: name of an argument used in this script
        :param arguments: argparse module instance with command line argument values
        :param arg_name: name of concrete argument from argparse
        :param check_func: what function should is used for value check of argument arg_name
        :param msg: what msg will be printed out if the value of argument is incorrect
    """
    value = getattr(arguments, arg_name)
    if check_func(value):
        args_dict[script_name] = value
    else:
        raise argparse.ArgumentTypeError(msg + value)


def resolve_parsed_names(arguments):
    """
        Helper function to make the use of environment variables and command line arguments easier in main.

        :param arguments: argparse module instance with command line argument values
        :return: Dictionary with converted command line arguments names and their values
    """
    args_dict = {}
    args_dict['LOCAL'] = str(getattr(arguments, 'local'))
    check_and_load_parsed_arg(args_dict, 'DATAWAREHOUSE', arguments, 'ip_and_port', is_ip_and_port,
                              'given string is not a valid <ip:port>: ')
    check_and_load_parsed_arg(args_dict, 'INPUT_DIR', arguments, 'input_dir', is_dir_path,
                              'given string is not a valid path to an existing directory: ')
    check_and_load_parsed_arg(args_dict, 'OUTPUT_DIR', arguments, 'output_dir', is_dir_path,
                              'given string is not a valid path to an existing directory: ')
    check_and_load_parsed_arg(args_dict, 'ZIP_DIR', arguments, 'zip_dir_path', is_abs_dir_path,
                              'given string is not a valid absolute path to an existing directory: ')
    check_and_load_parsed_arg(args_dict, 'RDF_FILE_NAME', arguments, 'rdf_file_name', is_string,
                              'given string is not a valid RDF file name: ')
    check_and_load_parsed_arg(args_dict, 'SCHEMA_FILE_NAME', arguments, 'schema_file_name', is_string,
                              'given string is not a valid SCHEMA file name: ')
    check_and_load_parsed_arg(args_dict, 'CASE_ID', arguments, 'case_id', is_string,
                              'given string is not a valid case_id: ')
    check_and_load_parsed_arg(args_dict, 'INPUT_IDENTIFIER', arguments, 'input_identifier', is_string,
                              'given string is not a valid input_identifier: ')
    return args_dict


if __name__ == '__main__':
    # TODO - if new log files are supposed to be transformed, add their names to log_file_names:
    log_file_names = ('conn', 'http', 'dns', 'ssl', 'notice', 'sip', 'x509', 'files', 'ftp',
                      'ssh', 'kerberos', 'smb_files', 'smb_mapping')

    args_loaded = define_parsed_arguments()

    if getattr(args_loaded, 'independent_mode'):
        args = resolve_parsed_names(args_loaded)
    else:
        prefix = get_prefix()
        args = define_env_arguments(prefix)
        # this environment variables prefix corresponds to the module name in orchestrator db

    local_storage = eval(args['LOCAL'])

    address = 'http://' + args["DATAWAREHOUSE"] + '/'
    logs_file_identifier = args['INPUT_IDENTIFIER']
    input_dir = args['INPUT_DIR']
    output_dir = args['OUTPUT_DIR']
    zip_dir = args['ZIP_DIR']
    case_id = args['CASE_ID']

    wc = warehouse_connector.WarehouseConnector(logs_file_identifier,
                                                input_dir,
                                                zip_dir,
                                                'transformed',
                                                True,
                                                'dgraph_transformation',
                                                'zeek_extraction',
                                                None,
                                                None,
                                                address)

    if not local_storage:
        case_id = wc.get_case_id()

    # arguments for the transformation script:
    script_arguments = get_input(local_storage, input_dir, logs_file_identifier, log_file_names, wc)

    if script_arguments:
        output_rdf_file = args['OUTPUT_DIR'] + args['RDF_FILE_NAME'] + '.rdf'
        output_schema_file = args['OUTPUT_DIR'] + args['SCHEMA_FILE_NAME'] + '.schema'
        
        try:
            # apply the transformation on log files in input_dir:
            subprocess.call(' python3 /usr/local/bin/granef/zeek-dgraph.py' + ' -l ' + input_dir + ' -or ' +
                            output_rdf_file + ' -os ' + output_schema_file + ' -s \'' + case_id + '\'', shell=True)

        except subprocess.CalledProcessError as e:
            print('ERROR (CalledProcessError) when running zeek-dgraph.py (errno=' + str(e.errno) + ', strerror' + 
                  str(e.strerror) + ', filename' + str(e.filename) + ').')
        except Exception as e:
            print('ERROR when running zeek-dgraph.py (' + str(e) + ').')

        if not local_storage:
            upload_zipped_output(output_dir, zip_dir, wc)
    else:
        print('No log files were loaded. No files will be uploaded.')
