#! /usr/bin/env python3
# -*- coding: utf-8 -*-

#
# Copyright (c) 2018-2019, Institute of Computer Science, Masaryk University
#
# Developers:
#   Milan Cermak <cermak@ics.muni.cz>
#   Denisa Sramkova <456424@mail.muni.cz>
#

"""
Module to parse dns.log file and transform it to the Dgraph database RDF format.
"""

import textwrap  # dedent() function to trim multiline strings
import ipaddress  # IP address processing
import validators  # IP and hostname validators
import hashlib  # Hash operations


def get_schema():
    """
    Definition of Dgraph schema for a domain names of given host and a dns log.

    Value host.domain is specified as a array of strings (behaves as set) and indexed as trigram to allow filtering
    by regular expressions.

    Value dns.query is indexed as trigram to allow filtering by regular expressions.

    :return: Defined schema as a string.
    """
    schema = """\
        dns.query: string @index(trigram, exact) .
        dns.qclass: int .
        dns.qclass_name: string .
        dns.qtype: int .
        dns.qtype_name: string .
        dns.aa: bool .
        dns.tc: bool .
        dns.rd: bool .
        dns.ra: bool .
        dns.z: int .
        dns.answer: string .
        dns.rejected: bool .
        host.hostname: [uid] @reverse .
        hostname.name: string @index(trigram, exact) .
        hostname.type: string @index(exact) .
    """
    return textwrap.dedent(schema)


def get_types():
    return [("Dns", ["dns.query",
                     "dns.qclass",
                     "dns.qclass_name",
                     "dns.qtype",
                     "dns.qtype_name",
                     "dns.aa",
                     "dns.tc",
                     "dns.rd",
                     "dns.ra",
                     "dns.z",
                     "dns.answer",
                     "dns.rejected"
                     # ,"<~connection.produced>"
                     ]),
            # ("Host", ["host.hostname"]),
            ("Hostname", ["hostname.name",
                          "hostname.type"
                          # ,"<~host.hostname>"
                         ])]


def resolve_dns_answers(id_dns, qtype_name, answers, query, source_name):
    """
    Auxiliary function for dns answer resolution.

    :param id_dns: Id of dns node
    :param qtype_name: Qtype value from dns Zeek log from record specified by id_dns
    :param answers: Answers value from dns Zeek log from record specified by id_dns
    :param query: Query value from dns Zeek log from record specified by id_dns
    :param source_name: Name of the data source
    :return: Resolved dns answers
    """
    if not answers:
        return

    answer_mutation = ""

    if qtype_name == "A" or qtype_name == "AAAA":  # IPv4, IPv6
        for answer in answers:
            try:
                ipaddress.ip_address(answer)
                answer_mutation += textwrap.dedent("""\
                    _:{id_dns} <dns.answer> "{answer}" .
                    _:{orig_h} <host.hostname> _:{hostname_id} .
                    _:{hostname_id} <hostname.name> "{hostname}" .
                    _:{hostname_id} <hostname.type> "dns" .
                    _:{hostname_id} <source> "{data_source}" .
                    _:{hostname_id} <dgraph.type> "Hostname" .
                """.format(
                    id_dns=id_dns,
                    answer=answer,
                    orig_h=answer.replace(':', '_0'),
                    hostname_id="hname-" + hashlib.sha1(query.replace(':', '_0').encode('utf-8')).hexdigest(),
                    hostname=query,
                    data_source=source_name
                ))
                break
            except ValueError:
                continue

    elif qtype_name == "NS":  # domain name
        for answer in answers:
            if validators.domain(answer):
                answer_mutation += """_:{id_dns} <dns.answer> "{answer}" .
                """.format(
                    id_dns=id_dns,
                    answer=answer
                )
                break

    else:
        answers_str = ''
        for answer_item in answers:
            answers_str += answer_item + ' '

        answer_mutation += textwrap.dedent("""\
            _:{id_dns} <dns.answer> "{answers}" .
        """.format(
            id_dns=id_dns,
            answers=hashlib.sha1(answers_str.replace('"', '').encode('utf-8')).hexdigest()
        ))

    return answer_mutation


def get_mutations(dns_json, source_name):
    """
    Parsing of the dns log file and definition of all host-domain name pairs in the nquads format.

    Data model: (IP)--connection.produced--dns-->[domain_names]

    :param dns_json: Loaded JSON with a dns log line from the Zeek log
    :param source_name: Name of the data source
    :return: Nquads definition for host and all his domain names
    """
    id_dns = "dns" + dns_json.get("uid", "").replace(':', '_0')
    qtype_name = dns_json.get("qtype_name", "")
    nquads_mutation = textwrap.dedent("""\
        _:{id_dns} <source> "{data_source}" .
        _:{id_dns} <dgraph.type> "Dns" .
        _:{id_dns} <dns.query> "{query}" .
        _:{id_dns} <dns.qclass> "{qclass}" .
        _:{id_dns} <dns.qclass_name> "{qclass_name}" .
        _:{id_dns} <dns.qtype> "{qtype}" .
        _:{id_dns} <dns.qtype_name> "{qtype_name}" .
        _:{id_dns} <dns.aa> "{aa}" .
        _:{id_dns} <dns.tc> "{tc}" .
        _:{id_dns} <dns.rd> "{rd}" .
        _:{id_dns} <dns.ra> "{ra}" .
        _:{id_dns} <dns.z> "{z}" .
        _:{id_dns} <dns.rejected> "{rejected}" .
        _:{connection_uid} <connection.produced> _:{id_dns} .
    """.format(
        data_source=source_name,
        id_dns=id_dns,
        query=dns_json.get("query", ""),
        qclass=dns_json.get("qclass", 0),
        qclass_name=dns_json.get("qclass_name", ""),
        qtype=dns_json.get("qtype", 0),
        qtype_name=qtype_name,
        aa=str(dns_json.get("AA", False)).lower(),
        tc=str(dns_json.get("TC", False)).lower(),
        rd=str(dns_json.get("RD", False)).lower(),
        ra=str(dns_json.get("RA", False)).lower(),
        z=dns_json.get("Z", 0),
        answers=dns_json.get("answers", ""),
        rejected=str(dns_json.get("rejected", False)).lower(),
        connection_uid=dns_json.get("uid", "").replace(':', '_0')
    ))

    if qtype_name:
        answers = resolve_dns_answers(id_dns, qtype_name, dns_json.get("answers", ""), dns_json.get("query", ""), source_name)
        if answers:
            nquads_mutation += answers

    return textwrap.dedent(nquads_mutation)
