#! /usr/bin/env python3
# -*- coding: utf-8 -*-

#
# Copyright (c) 2018-2019, Institute of Computer Science, Masaryk University
#
# Developers:
#   Milan Cermak <cermak@ics.muni.cz>
#   Denisa Sramkova <456424@mail.muni.cz>
#

"""
Module to parse files.log file and transform it to the Dgraph database RDF format.
"""

import textwrap  # dedent() function to trim multiline strings


def get_schema():
    """
    Definition of Dgraph schema for a files log.

    :return: Defined schema as a string.
    """
    schema = """\
        files.source: string .
        files.depth: int .
        files.analyzers: [string] .
        files.mime_type: string .
        files.duration: float .
        files.local_orig: bool .
        files.is_orig: bool .
        files.seen_bytes: int .
        files.total_bytes: int .
        files.missing_bytes: int .
        files.overflow_bytes: int .
        files.timedout: bool .
        files.fuid: [uid] @reverse .
        file.mime_type: string .
        file.total_bytes: int .
        file.md5: string @index(exact) .
        file.sha1: string @index(exact) .
        host.provided: [uid] @reverse .
        host.obtained: [uid] @reverse .
    """
    return textwrap.dedent(schema)


def get_types():
    return [("Files", ["files.source",
                       "files.depth",
                       "files.analyzers",
                       "files.mime_type",
                       "files.duration",
                       "files.local_orig",
                       "files.is_orig",
                       "files.seen_bytes",
                       "files.total_bytes",
                       "files.missing_bytes",
                       "files.overflow_bytes",
                       "files.timedout",
                       # ,"files.fuid"
                       # ,"<~connection.produced>"
                       ]),
            # ("Host", ["host.provided", "host.obtained"]),
            ("File", ["file.mime_type",
                      "file.total_bytes",
                      "file.md5",
                      "file.sha1"
                      # ,"<~files.fuid>",
                      # "<~host.provided>",
                      # "<~host.obtained>"
                      ])]


def get_mutations(files_json, source_name):
    """
    Parsing of the files log file with the difinition of file as a distinct node.

    :param files_json: Loaded JSON with a files log line from the Zeek log
    :param source_name: Name of the data source
    :return: Full nquads definition for given files node
    """
    id_file = files_json.get("fuid", "").replace(':', '_0')
    nquads_mutation = textwrap.dedent("""\
        _:{id_file} <source> "{data_source}" .
        _:{id_file} <dgraph.type> "Files" .
        _:{id_file} <files.source> "{source}" .
        _:{id_file} <files.depth> "{depth}" .
        _:{id_file} <files.analyzers> "{analyzers}" .
        _:{id_file} <files.mime_types> "{mime_types}" .
        _:{id_file} <files.duration> "{duration}" .
        _:{id_file} <files.local_orig> "{local_orig}" .
        _:{id_file} <files.is_orig> "{is_orig}" .
        _:{id_file} <files.seen_bytes> "{seen_bytes}" .
        _:{id_file} <files.total_bytes> "{total_bytes}" .
        _:{id_file} <files.missing_bytes> "{missing_bytes}" .
        _:{id_file} <files.overflow_bytes> "{overflow_bytes}" .
        _:{id_file} <files.timedout> "{timedout}" .
        _:{id_file} <files.fuid> _:{id_file}-file .
    """.format(
        data_source=source_name,
        id_file=id_file,
        source=files_json.get("source", ""),
        depth=files_json.get("depth", 0),
        analyzers=files_json.get("analyzers", ""),
        mime_types=files_json.get("mime_types", ""),
        duration=files_json.get("duration", 0.0),
        local_orig=str(files_json.get("local_orig", False)).lower(),
        is_orig=str(files_json.get("is_orig", False)).lower(),
        seen_bytes=files_json.get("seen_bytes", 0),
        total_bytes=files_json.get("total_bytes", 0),
        missing_bytes=files_json.get("seen_bytes", 0),
        overflow_bytes=files_json.get("overflow_bytes", 0),
        timedout=str(files_json.get("timedout", False)).lower(),
        md5=files_json.get("md5", ""),
        sha1=files_json.get("sha1", "")
    ))

    for connection_uid in files_json.get("conn_uids", ""):
        nquads_mutation += textwrap.dedent("""\
            _:{connection_uid} <connection.produced> _:{id_files} .
        """.format(
            id_files=id_file,
            connection_uid=connection_uid.replace(':', '_0')
        ))

    # Append the file to host
    nquads_mutation += textwrap.dedent("""\
        _:{id_file}-file <source> "{data_source}" .
        _:{id_file}-file <dgraph.type> "File" .
        _:{id_file}-file <file.mime_type> "{mime_type}" .
        _:{id_file}-file <file.total_bytes> "{total_bytes}" .
        _:{id_file}-file <file.md5> "{md5}" .
        _:{id_file}-file <file.sha1> "{sha1}" .
    """.format(
        data_source=source_name,
        id_file=id_file,
        mime_type=files_json.get("mime_type", ""),
        total_bytes=files_json.get("total_bytes", 0),
        md5=files_json.get("md5", ""),
        sha1=files_json.get("sha1", "")
    ))

    for source_host in files_json.get("tx_hosts", []):
        nquads_mutation += textwrap.dedent("""\
            _:{orig_h} <host.provided> _:{id_file}-file  .
        """.format(
            orig_h=source_host.replace(':', '_0'),
            id_file=id_file
        ))
    for destination_host in files_json.get("rx_hosts", []):
        nquads_mutation += textwrap.dedent("""\
            _:{orig_h} <host.obtained> _:{id_file}-file  .
        """.format(
            orig_h=destination_host.replace(':', '_0'),
            id_file=id_file
        ))

    return textwrap.dedent(nquads_mutation)
