#! /usr/bin/env python3
# -*- coding: utf-8 -*-

#
# Copyright (c) 2018-2019, Institute of Computer Science, Masaryk University
#
# Developers:
#   Milan Cermak <cermak@ics.muni.cz>
#   Denisa Sramkova <456424@mail.muni.cz>
#

"""
Module to parse ftp.log file and transform it to the Dgraph database RDF format.
"""

import textwrap  # dedent() function to trim multiline strings


def get_schema():
    """
    Definition of Dgraph schema for a ftp log.

    :return: Defined schema as a string.
    """
    schema = """\
        ftp.user: string .
        ftp.password: string .
        ftp.command: string .
        ftp.arg: string .
        ftp.mime_type: string .
        ftp.reply_code: int .
        ftp.reply_msg: string .
        ftp.data_channel_passive: bool .
        ftp.data_channel_orig_h: string .
        ftp.data_channel_resp_h: string .
        ftp.data_channel_resp_p: int .
        ftp.fuid: [uid] @reverse .
        file.name: [string] .
    """
    return textwrap.dedent(schema)


def get_types():
    return [("Ftp", ["ftp.user",
                     "ftp.password",
                     "ftp.command",
                     "ftp.arg",
                     "ftp.mime_type",
                     "ftp.reply_code",
                     "ftp.reply_msg",
                     "ftp.data_channel_passive",
                     "ftp.data_channel_orig_h",
                     "ftp.data_channel_resp_h",
                     "ftp.data_channel_resp_p"
                     # ,"ftp.fuid"
                     # ,"<~connection.produced>"
                     ]),
            ("File", ["file.name",
                      # "<~ftp.fuid>"
                      ])]


def get_mutations(ftp_json, source_name):
    """
    Parsing of the ftp log file with the definition of file as a distinct node.

    :param ftp_json: Loaded JSON with a ftp log line from the Zeek log
    :param source_name: Name of the data source
    :return: Full nquads definition for given ftp node
    """
    id_ftp = "ftp" + ftp_json.get("uid", "").replace(':', '_0')
    nquads_mutation = textwrap.dedent("""\
        _:{id_ftp} <source> "{data_source}" .
        _:{id_ftp} <type> "ftp" .
        _:{id_ftp} <ftp.user> "{user}" .
        _:{id_ftp} <ftp.password> "{password}" .
        _:{id_ftp} <ftp.command> "{command}" .
        _:{id_ftp} <ftp.arg> "{arg}" .
        _:{id_ftp} <ftp.mime_type> "{mime_type}" .
        _:{id_ftp} <ftp.reply_code> "{reply_code}" .
        _:{id_ftp} <ftp.reply_msg> "{reply_msg}" .
        _:{id_ftp} <ftp.data_channel_passive> "{data_channel_passive}" .
        _:{id_ftp} <ftp.data_channel_orig_h> "{data_channel_orig_h}" .
        _:{id_ftp} <ftp.data_channel_resp_h> "{data_channel_resp_h}" .
        _:{id_ftp} <ftp.data_channel_resp_p> "{data_channel_resp_p}" .
        _:{connection_uid} <connection.produced> _:{id_ftp} .
    """.format(
        data_source=source_name,
        id_ftp=id_ftp,
        user=ftp_json.get("user", ""),
        password=ftp_json.get("password", ""),
        command=ftp_json.get("command", ""),
        arg=ftp_json.get("arg", ""),
        mime_type=ftp_json.get("mime_type", ""),
        reply_code=ftp_json.get("reply_code", 0),
        reply_msg=ftp_json.get("reply_msg", ""),
        data_channel_passive=str(ftp_json.get("data_channel.passive", False)).lower(),
        data_channel_orig_h=ftp_json.get("data_channel.orig_h", ""),
        data_channel_resp_h=ftp_json.get("data_channel.resp_h", ""),
        data_channel_resp_p=ftp_json.get("data_channel.resp_p", 0),
        connection_uid=ftp_json.get("uid", "").replace(':', '_0')
    ))

    if ftp_json.get("fuid", ""):
        nquads_mutation += textwrap.dedent("""\
            _:{id_ftp} <ftp.fuid> _:{fuid} .
            _:{fuid}-file <file.name> "{arg}" .
            _:{fuid}-file <source> "{data_source}" .
            _:{fuid}-file <type> "file" .
        """.format(
            id_ftp=id_ftp,
            fuid=ftp_json.get("fuid", "").replace(':', '_0'),
            arg=ftp_json.get("arg", ""),
            data_source=source_name
        ))

    return textwrap.dedent(nquads_mutation)
