#! /usr/bin/env python3
# -*- coding: utf-8 -*-

#
# Copyright (c) 2018-2019, Institute of Computer Science, Masaryk University
#
# Developers:
#   Milan Cermak <cermak@ics.muni.cz>
#   Denisa Sramkova <456424@mail.muni.cz>
#

"""
Module to parse smb_mapping.log file and transform it to the Dgraph database RDF format.
"""

import textwrap  # dedent() function to trim multiline strings


def get_schema():
    """
    Definition of Dgraph schema for a smb_mapping log.

    :return: Defined schema as a string.
    """
    schema = """\
        smb_mapping.path: string .
        smb_mapping.service: string .
        smb_mapping.share_type: string .
    """
    return textwrap.dedent(schema)


def get_types():
    return [("Smb_mapping", ["smb_mapping.path",
                             "smb_mapping.service",
                             "smb_mapping.share_type"
                             # ,"<~connection.produced>"
                             ])]


def get_mutations(smb_mapping_json, source_name):
    """
    Parsing of the smb_mapping log file with the difinition of file as a distinct node.

    :param smb_mapping_json: Loaded JSON with a smb_mapping log line from the Zeek log
    :param source_name: Name of the data source
    :return: Full nquads definition for given smb_mapping node
    """
    id_smb_mapping = "smbm" + smb_mapping_json.get("id", "").replace(':', '_0')
    nquads_mutation = textwrap.dedent("""\
        _:{id_smb_mapping} <source> "{data_source}" .
        _:{id_smb_mapping} <dgraph.type> "Smb_mapping" .
        _:{id_smb_mapping} <smb_mapping.path> "{path}" .
        _:{id_smb_mapping} <smb_mapping.service> "{service}" .
        _:{id_smb_mapping} <smb_mapping.share_type> "{share_type}" .
        _:{connection_uid} <connection.produced> _:{id_smb_mapping} .
    """.format(
        data_source=source_name,
        id_smb_mapping=id_smb_mapping,
        path=smb_mapping_json.get("path", "").replace('\\', '\\\\'),
        service=smb_mapping_json.get("service", ""),
        share_type=smb_mapping_json.get("share_type", ""),
        connection_uid=smb_mapping_json.get("uid", "").replace(':', '_0')
    ))

    return textwrap.dedent(nquads_mutation)
