#! /usr/bin/env python3
# -*- coding: utf-8 -*-

#
# Copyright (c) 2018-2019, Institute of Computer Science, Masaryk University
#
# Developers:
#   Milan Cermak <cermak@ics.muni.cz>
#   Denisa Sramkova <456424@mail.muni.cz>
#

"""
Module to parse ssl.log file and transform it to the Dgraph database RDF format.
"""

import textwrap  # dedent() function to trim multiline strings
import hashlib  # Hash operations


def get_schema():
    """
    Definition of Dgraph schema for a ssl log.

    :return: Defined schema as a string.
    """
    schema = """\
        ssl.version: string .
        ssl.cipher: string .
        ssl.resumed: bool .
        ssl.established: bool .
        ssl.cert_chain_fuid: [uid] @reverse .
        host.x509: [uid] @reverse .
        x509.certificate_subject: string .
        x509.certificate_issuer: string .
        host.hostname: [uid] @reverse .
        hostname.name: string @index(trigram, exact) .
        hostname.type: string @index(exact) .
        ssl.hostname: [uid] @reverse .
    """
    return textwrap.dedent(schema)


def get_types():
    return [("Ssl", [# "ssl.hostname",
                     # "ssl.cert_chain_fuid",
                     "ssl.version",
                     "ssl.cipher",
                     "ssl.resumed",
                     "ssl.established"
                     # ,"<~connection.produced>"
                     ]),
            # ("Host", ["host.hostname","host.x509"]),
            ("Hostname", ["hostname.name",
                          "hostname.type"
                          # ,"<~host.hostname>"
                          # "<~ssl.hostname>"
                          ]),
            ("X509", ["x509.certificate_subject",
                      "x509.certificate_issuer"
                      # "<~ssl.cert_chain_fuid>",
                      # "<~host.x509>"
                      ])]


def get_mutations(ssl_json, source_name):
    """
    Parsing of the ssl log file and definition of x509 and hostname as distinct nodes.

    :param ssl_json: Loaded JSON with a ssl log line from the Zeek log
    :param source_name: Name of the data source
    :return: Full nquads definition for given ssl node
    """
    id_ssl = "ssl" + ssl_json.get("uid", "").replace(':', '_0')
    nquads_mutation = textwrap.dedent("""\
        _:{id_ssl} <source> "{data_source}" .
        _:{id_ssl} <dgraph.type> "Ssl" .
        _:{id_ssl} <ssl.version> "{version}" .
        _:{id_ssl} <ssl.cipher> "{cipher}" .
        _:{id_ssl} <ssl.resumed> "{resumed}" .
        _:{id_ssl} <ssl.established> "{established}" .
        _:{connection_uid} <connection.produced> _:{id_ssl} .
    """.format(
        data_source=source_name,
        id_ssl=id_ssl,
        version=ssl_json.get("version", ""),
        cipher=ssl_json.get("cipher", ""),
        resumed=str(ssl_json.get("resumed", False)).lower(),
        established=str(ssl_json.get("established", False)).lower(),
        connection_uid=ssl_json.get("uid", "").replace(':', '_0')
    ))

    if "subject" in ssl_json or "issuer" in ssl_json:
        for fuid in ssl_json.get("cert_chain_fuids", ""):
            nquads_mutation += textwrap.dedent("""\
                _:{id_ssl} <ssl.cert_chain_fuid> _:{fuid} .
                _:{host} <host.x509> _:{fuid} .
                _:{fuid} <x509.certificate_subject> "{certificate_subject}" .
                _:{fuid} <x509.certificate_issuer> "{certificate_issuer}" .
                _:{fuid} <source> "{data_source}" .
                _:{fuid} <dgraph.type> "X509" .
            """.format(
                id_ssl=id_ssl,
                fuid="cert-"+fuid.replace(':', '_0'),
                host=ssl_json.get("id.resp_h", "").replace(':', '_0'),
                certificate_subject=ssl_json.get("subject", "").replace('"', '').replace('\\', '-'),
                certificate_issuer=ssl_json.get("issuer", "").replace('\\', '-'),
                data_source=source_name
            ))

    # Add hostname to the host
    if "server_name" in ssl_json:
        nquads_mutation += textwrap.dedent("""\
            _:{host} <host.hostname> _:{hostname_id} .
            _:{hostname_id} <hostname.name> "{hostname}" .
            _:{hostname_id} <hostname.type> "ssl" .
            _:{hostname_id} <source> "{data_source}" .
            _:{hostname_id} <dgraph.type> "Hostname" .
            _:{id_ssl} <ssl.hostname> _:{hostname_id} .
        """.format(
            host=ssl_json.get("id.resp_h", "").replace(':', '_0'),
            hostname_id="hname-" + hashlib.sha1(ssl_json.get("server_name", "").replace(':', '_0').encode('utf-8')).hexdigest(),
            hostname=ssl_json.get("server_name", ""),
            id_ssl=id_ssl,
            data_source=source_name
        ))

    return textwrap.dedent(nquads_mutation)
